module libav.libavfilter.avfilter;
/*
 * filter layer
 * Copyright (c) 2007 Bobby Bingham*
 
 * This file is part of Libav.*
 
 * Libav is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.*
 
 * Libav is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.*
 
 * You should have received a copy of the GNU Lesser General Public
 * License along with Libav; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

/**
 * @file
 * @ingroup lavfi
 * Main libavfilter public API header
 */

/**
 * @defgroup lavfi Libavfilter - graph-based frame editing library
 * @{
 */

import libav.libavutil.avutil;
import libav.libavutil.dict;
import libav.libavutil.frame;
import libav.libavutil.log;
import libav.libavutil.samplefmt;
import libav.libavutil.pixfmt;
import libav.libavutil.rational;
import libav.libavcodec.avcodec;
import libav.libavformat.avformat;

import libav.libavfilter._version;

extern (C):

/**
 * Return the LIBAVFILTER_VERSION_INT constant.
 */
uint avfilter_version();

/**
 * Return the libavfilter build-time configuration.
 */
char* avfilter_configuration();

/**
 * Return the libavfilter license.
 */
char* avfilter_license();

struct AVFilterFormats{}
struct AVFilterChannelLayouts{};

static if(FF_API_AVFILTERBUFFER){
    /**
     * A reference-counted buffer data type used by the filter system. Filters
     * should not store pointers to this structure directly, but instead use the
     * AVFilterBufferRef structure below.
     */
    struct AVFilterBuffer {
        ubyte*[8] data;           ///< buffer data for each plane/channel

        /**
         * pointers to the data planes/channels.*
         
         * For video, this should simply point to data[].*
         
         * For planar audio, each channel has a separate data pointer, and
         * linesize[0] contains the size of each channel buffer.
         * For packed audio, there is just one data pointer, and[0] linesize
         * contains the total size of the buffer for all channels.*
         
         * Note: Both data and extended_data will always be set, but for planar
         * audio with more channels that can fit in data, extended_data must be used
         * in order to access all channels.
         */
        ubyte** extended_data;
        int[8] linesize;            ///< number of bytes per line

        /** private data to be used by a custom free function */
        void* priv;
        /**
         * A pointer to the function to deallocate this buffer if the default
         * function is not sufficient. This could, for example, add the memory
         * back into a memory pool to be reused later without the overhead of
         * reallocating it from scratch.
         */
        void function(AVFilterBuffer* buf) free;

        int format;             ///< media format
        int w, h;               ///< width and height of the allocated buffer
        uint refcount;          ///< number of references to this buffer
    }
}else{
    struct AVFilterBuffer{}
}

immutable int AV_PERM_READ          = 0x01;  ///< can read from the buffer
immutable int AV_PERM_WRITE         = 0x02;  ///< can write to the buffer
immutable int AV_PERM_PRESERVE      = 0x04;  ///< nobody else can overwrite the buffer
immutable int AV_PERM_REUSE         = 0x08;  ///< can output the buffer multiple times, with the same contents each time
immutable int AV_PERM_REUSE2        = 0x10;  ///< can output the buffer multiple times, modified each time
immutable int AV_PERM_NEG_LINESIZES = 0x20;  ///< the buffer requested can have negative linesizes

/**
 * Audio specific properties in a reference to an AVFilterBuffer. Since
 * AVFilterBufferRef is common to different media formats, audio specific
 * per reference properties must be separated out.
 */
struct AVFilterBufferRefAudioProps {
    ulong channel_layout;    ///< channel layout of audio buffer
    int nb_samples;             ///< number of audio samples
    int sample_rate;            ///< audio buffer sample rate
    int planar;                 ///< audio buffer - planar or packed
}

/**
 * Video specific properties in a reference to an AVFilterBuffer. Since
 * AVFilterBufferRef is common to different media formats, video specific
 * per reference properties must be separated out.
 */
struct AVFilterBufferRefVideoProps {
    int w;                      ///< image width
    int h;                      ///< image height
    AVRational pixel_aspect;    ///< pixel aspect ratio
    int interlaced;             ///< is frame interlaced
    int top_field_first;        ///< field order
    AVPictureType pict_type;    ///< picture type of the frame
    int key_frame;              ///< 1 -> keyframe, 0-> not
}

/**
 * A reference to an AVFilterBuffer. Since filters can manipulate the origin of
 * a buffer to, for example, crop image without any memcpy, the buffer origin
 * and dimensions are per-reference properties. Linesize is also useful for
 * image flipping, frame to field filters, etc, and so is also per-reference.*
 
 * TODO: add anything necessary for frame reordering
 */
struct AVFilterBufferRef {
    AVFilterBuffer* buf;        ///< the buffer that this is a reference to
    ubyte*[8] data;           ///< picture/audio data for each plane
    /**
     * pointers to the data planes/channels.*
     
     * For video, this should simply point to data[].*
     
     * For planar audio, each channel has a separate data pointer, and
     * linesize[0] contains the size of each channel buffer.
     * For packed audio, there is just one data pointer, and[0] linesize
     * contains the total size of the buffer for all channels.*
     
     * Note: Both data and extended_data will always be set, but for planar
     * audio with more channels that can fit in data, extended_data must be used
     * in order to access all channels.
     */
    ubyte** extended_data;
    int[8] linesize;            ///< number of bytes per line

    AVFilterBufferRefVideoProps* video; ///< video buffer specific properties
    AVFilterBufferRefAudioProps* audio; ///< audio buffer specific properties

    /**
     * presentation timestamp. The time unit may change during
     * filtering, as it is specified in the link and the filter code
     * may need to rescale the PTS accordingly.
     */
    long pts;
    long pos;                ///< byte position in stream, -1 if unknown

    int format;                 ///< media format

    int perms;                  ///< permissions, see the AV_PERM_* flags

    AVMediaType type;           ///< media type of buffer data
}

/**
 * Copy properties of src to dst, without copying the actual data
 */
void avfilter_copy_buffer_ref_props(AVFilterBufferRef* dst, AVFilterBufferRef* src);

/**
 * Add a new reference to a buffer.*
 
 * @param ref   an existing reference to the buffer
 * @param pmask a bitmask containing the allowable permissions in the new
 *              reference
 * @return      a new reference to the buffer with the same properties as the
 *              old, excluding any permissions denied by pmask
 */
AVFilterBufferRef* avfilter_ref_buffer(AVFilterBufferRef* _ref, int pmask);

/**
 * Remove a reference to a buffer. If this is the last reference to the
 * buffer, the buffer itself is also automatically freed.*
 
 * @param ref reference to the buffer, may be NULL*
 
 * @note it is recommended to use avfilter_unref_bufferp() instead of this
 * function
 */
void avfilter_unref_buffer(AVFilterBufferRef* _ref);

/**
 * Remove a reference to a buffer and set the pointer to NULL.
 * If this is the last reference to the buffer, the buffer itself
 * is also automatically freed.*
 
 * @param ref pointer to the buffer reference
 */
void avfilter_unref_bufferp(AVFilterBufferRef** _ref);

static if(FF_API_AVFILTERPAD_PUBLIC){
    /**
     * A filter pad used for either input or output.*
     
     * @warning this struct will be removed from public API.
     * users should call avfilter_pad_get_name() and avfilter_pad_get_type()
     * to access the name and type fields; there should be no need to access
     * any other fields from outside of libavfilter.
     */
    struct AVFilterPad {
        /**
         * Pad name. The name is unique among inputs and among outputs, but an
         * input may have the same name as an output. This may be NULL if this
         * pad has no need to ever be referenced by name.
         */
        char* name;

        /**
         * AVFilterPad type.
         */
        AVMediaType type;

        /**
         * Minimum required permissions on incoming buffers. Any buffer with
         * insufficient permissions will be automatically copied by the filter
         * system to a new buffer which provides the needed access permissions.*
         
         * Input pads only.
         */
        int min_perms;

        /**
         * Permissions which are not accepted on incoming buffers. Any buffer
         * which has any of these permissions set will be automatically copied
         * by the filter system to a new buffer which does not have those
         * permissions. This can be used to easily disallow buffers with
         * AV_PERM_REUSE.*
         
         * Input pads only.
         */
        int rej_perms;

        /**
         * @deprecated unused
         */
        int function(AVFilterLink* link, AVFilterBufferRef* picref) start_frame;

        /**
         * Callback function to get a video buffer. If NULL, the filter system will
         * use avfilter_default_get_video_buffer().*
         
         * Input video pads only.
         */
        AVFrame* function(AVFilterLink* link, int w, int h) get_video_buffer;

        /**
         * Callback function to get an audio buffer. If NULL, the filter system will
         * use avfilter_default_get_audio_buffer().*
         
         * Input audio pads only.
         */
        AVFrame* function(AVFilterLink* link, int nb_samples) get_audio_buffer;

        /**
         * @deprecated unused
         */
        int function(AVFilterLink* link) end_frame;

        /**
         * @deprecated unused
         */
        int function(AVFilterLink* link, int y, int height, int slice_dir) draw_slice;

        /**
         * Filtering callback. This is where a filter receives a frame with
         * audio/video data and should do its processing.*
         
         * Input pads only.*
         
         * @return >= 0 on success, a negative AVERROR on error. This function
         * must ensure that samplesref is properly unreferenced on error if it
         * hasn't been passed on to another filter.
         */
        int function(AVFilterLink* link, AVFrame* frame) filter_frame;

        /**
         * Frame poll callback. This returns the number of immediately available
         * samples. It should return a positive value if the next request_frame()
         * is guaranteed to return one frame (with no delay).*
         
         * Defaults to just calling the source poll_frame() method.*
         
         * Output pads only.
         */
        int function(AVFilterLink* link) poll_frame;

        /**
         * Frame request callback. A call to this should result in at least one
         * frame being output over the given link. This should return zero on
         * success, and another value on error.*
         
         * Output pads only.
         */
        int function(AVFilterLink* link) request_frame;

        /**
         * Link configuration callback.*
         
         * For output pads, this should set the link properties such as
         * width/height. This should NOT set the format property - that is
         * negotiated between filters by the filter system using the
         * query_formats() callback before this function is called.*
         
         * For input pads, this should check the properties of the link, and update
         * the filter's internal state as necessary.*
         
         * For both input and output filters, this should return zero on success,
         * and another value on error.
         */
        int function(AVFilterLink* link) config_props;

        /**
         * The filter expects a fifo to be inserted on its input link,
         * typically because it has a delay.*
         
         * input pads only.
         */
        int needs_fifo;

        int needs_writable;
    }
}else{
    struct AVFilterPad{}
}

/**
 * Get the number of elements in a NULL-terminated array of AVFilterPads (e.g.
 * AVFilter.inputs/outputs).
 */
int avfilter_pad_count(AVFilterPad* pads);

/**
 * Get the name of an AVFilterPad.*
 
 * @param pads an array of AVFilterPads
 * @param pad_idx index of the pad in the array it; is the caller's
 *                responsibility to ensure the index is valid*
 
 * @return name of the pad_idx'th pad in pads
 */
char* avfilter_pad_get_name(AVFilterPad* pads, int pad_idx);

/**
 * Get the type of an AVFilterPad.*
 
 * @param pads an array of AVFilterPads
 * @param pad_idx index of the pad in the array; it is the caller's
 *                responsibility to ensure the index is valid*
 
 * @return type of the pad_idx'th pad in pads
 */
AVMediaType avfilter_pad_get_type(AVFilterPad* pads, int pad_idx);

/**
 * The number of the filter inputs is not determined just by AVFilter.inputs.
 * The filter might add additional inputs during initialization depending on the
 * options supplied to it.
 */
immutable int AVFILTER_FLAG_DYNAMIC_INPUTS      = (1 << 0);
/**
 * The number of the filter outputs is not determined just by AVFilter.outputs.
 * The filter might add additional outputs during initialization depending on
 * the options supplied to it.
 */
immutable int AVFILTER_FLAG_DYNAMIC_OUTPUTS     = (1 << 1);
/**
 * The filter supports multithreading by splitting frames into multiple parts
 * and processing them concurrently.
 */
immutable int AVFILTER_FLAG_SLICE_THREADS       = (1 << 2);

/**
 * Filter definition. This defines the pads a filter contains, and all the
 * callback functions used to interact with the filter.
 */
struct AVFilter {
    /**
     * Filter name. Must be non-NULL and unique among filters.
     */
    char* name;

    /**
     * A description of the filter. May be NULL.*
     
     * You should use the NULL_IF_CONFIG_SMALL() macro to define it.
     */
    char* description;

    /**
     * List of inputs, terminated by a zeroed element.*
     
     * NULL if there are no (static) inputs. Instances of filters with
     * AVFILTER_FLAG_DYNAMIC_INPUTS set may have more inputs than present in
     * this list.
     */
    AVFilterPad* inputs;
    /**
     * List of outputs, terminated by a zeroed element.*
     
     * NULL if there are no (static) outputs. Instances of filters with
     * AVFILTER_FLAG_DYNAMIC_OUTPUTS set may have more outputs than present in
     * this list.
     */
    AVFilterPad* outputs;

    /**
     * A class for the private data, used to declare filter private AVOptions.
     * This field is NULL for filters that do not declare any options.*
     
     * If this field is non-NULL, the first member of the filter private data
     * must be a pointer to AVClass, which will be set by libavfilter generic
     * code to this class.
     */
    AVClass* priv_class;

    /**
     * A combination of AVFILTER_FLAG_*
     */
    int flags;

    /*****************************************************************
     * All fields below this line are not part of the public API. They
     * may not be used outside of libavfilter and can be changed and
     * removed at will.
     * New public fields should be added right above.*****************************************************************
     
     */

    /**
     * Filter initialization function.*
     
     * This callback will be called only once during the filter lifetime, after
     * all the options have been set, but before links between filters are
     * established and format negotiation is done.*
     
     * Basic filter initialization should be done here. Filters with dynamic
     * inputs and/or outputs should create those inputs/outputs here based on
     * provided options. No more changes to this filter's inputs/outputs can be
     * done after this callback.*
     
     * This callback must not assume that the filter links exist or frame
     * parameters are known.*
     
     * @ref AVFilter.uninit "uninit" is guaranteed to be called even if
     * initialization fails, so this callback does not have to clean up on
     * failure.*
     
     * @return 0 on success, a negative AVERROR on failure
     */
    int function(AVFilterContext* ctx) init;

    /**
     * Should be set instead of @ref AVFilter.init "init" by the filters that
     * want to pass a dictionary of AVOptions to nested contexts that are
     * allocated during init.*
     
     * On return, the options dict should be freed and replaced with one that
     * contains all the options which could not be processed by this filter (or
     * with NULL if all the options were processed).*
     
     * Otherwise the semantics is the same as for @ref AVFilter.init "init".
     */
    int function(AVFilterContext* ctx, AVDictionary** options) init_dict;

    /**
     * Filter uninitialization function.*
     
     * Called only once right before the filter is freed. Should deallocate any
     * memory held by the filter, release any buffer references, etc. It does
     * not need to deallocate the AVFilterContext.priv memory itself.*
     
     * This callback may be called even if @ref AVFilter.init "init" was not
     * called or failed, so it must be prepared to handle such a situation.
     */
    void function(AVFilterContext* ctx) uninit;

    /**
     * Query formats supported by the filter on its inputs and outputs.*
     
     * This callback is called after the filter is initialized (so the inputs
     * and outputs are fixed), shortly before the format negotiation. This
     * callback may be called more than once.*
     
     * This callback must set AVFilterLink.out_formats on every input link and
     * AVFilterLink.in_formats on every output link to a list of pixel/sample
     * formats that the filter supports on that link. For audio links, this
     * filter must also set @ref AVFilterLink.in_samplerates "in_samplerates" /
     * @ref AVFilterLink.out_samplerates "out_samplerates" and
     * @ref AVFilterLink.in_channel_layouts "in_channel_layouts" /
     * @ref AVFilterLink.out_channel_layouts "out_channel_layouts" analogously.*
     
     * This callback may be NULL for filters with one input, in which case
     * libavfilter assumes that it supports all input formats and preserves
     * them on output.*
     
     * @return zero on success, a negative value corresponding to an
     * AVERROR code otherwise
     */
    int function(AVFilterContext* ) query_formats;

    int priv_size;      ///< size of private data to allocate for the filter

    /**
     * Used by the filter registration system. Must not be touched by any other
     * code.
     */
    AVFilter* next;
}

/**
 * Process multiple parts of the frame concurrently.
 */
immutable int AVFILTER_THREAD_SLICE = (1 << 0);

struct AVFilterInternal{}

/** An instance of a filter */
struct AVFilterContext {
    AVClass* av_class;              ///< needed for av_log()

    AVFilter* filter;         ///< the AVFilter of which this is an instance

    char* name;                     ///< name of this filter instance

    AVFilterPad*   input_pads;      ///< array of input pads
    AVFilterLink** inputs;          ///< array of pointers to input links
    static if(FF_API_FOO_COUNT){
        uint input_count;           ///< @deprecated use nb_inputs
        uint    nb_inputs;          ///< number of input pads
    }
        AVFilterPad*   output_pads;     ///< array of output pads
        AVFilterLink** outputs;         ///< array of pointers to output links
    static if(FF_API_FOO_COUNT){
        uint output_count;          ///< @deprecated use nb_outputs
    }
    uint     nb_outputs;            ///< number of output pads
    void* priv;                     ///< private data for use by the filter

    AVFilterGraph* graph;          ///< filtergraph this filter belongs to

    /**
     * Type of multithreading being allowed/used. A combination of
     * AVFILTER_THREAD_* flags.*
     
     * May be set by the caller before initializing the filter to forbid some
     * or all kinds of multithreading for this filter. The default is allowing
     * everything.*
     
     * When the filter is initialized, this field is combined using bit AND with
     * AVFilterGraph.thread_type to get the final mask used for determining
     * allowed threading types. I.e. a threading type needs to be set in both
     * to be allowed.*
     
     * After the filter is initialzed, libavfilter sets this field to the
     * threading type that is actually used (0 for no multithreading).
     */
    int thread_type;

    /**
     * An opaque struct for libavfilter internal use.
     */
    AVFilterInternal* internal;
}

/**
 * A link between two filters. This contains pointers to the source and
 * destination filters between which this link exists, and the indexes of
 * the pads involved. In addition, this link also contains the parameters
 * which have been negotiated and agreed upon between the filter, such as
 * image dimensions, format, etc.
 */
struct AVFilterLink {
    AVFilterContext* src;       ///< source filter
    AVFilterPad* srcpad;        ///< output pad on the source filter

    AVFilterContext* dst;       ///< dest filter
    AVFilterPad* dstpad;        ///< input pad on the dest filter

    AVMediaType type;      ///< filter media type

    /* These parameters apply only to video */
    int w;                      ///< agreed upon image width
    int h;                      ///< agreed upon image height
    AVRational sample_aspect_ratio; ///< agreed upon sample aspect ratio
    /* These two parameters apply only to audio */
    ulong channel_layout;    ///< channel layout of current buffer (see libavutil/channel_layout.h)
    int sample_rate;            ///< samples per second

    int format;                 ///< agreed upon media format

    /**
     * Define the time base used by the PTS of the frames/samples
     * which will pass through this link.
     * During the configuration stage, each filter is supposed to
     * change only the output timebase, while the timebase of the
     * input link is assumed to be an unchangeable property.
     */
    AVRational time_base;

    /*****************************************************************
     * All fields below this line are not part of the public API. They
     * may not be used outside of libavfilter and can be changed and
     * removed at will.
     * New public fields should be added right above.*****************************************************************
     
     */
    /**
     * Lists of formats supported by the input and output filters respectively.
     * These lists are used for negotiating the format to actually be used,
     * which will be loaded into the format member, above, when chosen.
     */
    AVFilterFormats* in_formats;
    AVFilterFormats* out_formats;

    /**
     * Lists of channel layouts and sample rates used for automatic
     * negotiation.
     */
    AVFilterFormats*  in_samplerates;
    AVFilterFormats* out_samplerates;
    AVFilterChannelLayouts*  in_channel_layouts;
    AVFilterChannelLayouts* out_channel_layouts;

    /**
     * Audio only, the destination filter sets this to a non-zero value to
     * request that buffers with the given number of samples should be sent to
     * it. AVFilterPad.needs_fifo must also be set on the corresponding input
     * pad.
     * Last buffer before EOF will be padded with silence.
     */
    int request_samples;

    /** stage of the initialization of the link properties (dimensions, etc) */
    enum init_state{
        AVLINK_UNINIT = 0,      ///< not started
        AVLINK_STARTINIT,       ///< started, but incomplete
        AVLINK_INIT             ///< complete
    }
}

/**
 * Link two filters together.*
 
 * @param src    the source filter
 * @param srcpad index of the output pad on the source filter
 * @param dst    the destination filter
 * @param dstpad index of the input pad on the destination filter
 * @return       zero on success
 */
int avfilter_link(AVFilterContext* src, uint srcpad,
                  AVFilterContext* dst, uint dstpad);

/**
 * Negotiate the media format, dimensions, etc of all inputs to a filter.*
 
 * @param filter the filter to negotiate the properties for its inputs
 * @return       zero on successful negotiation
 */
int avfilter_config_links(AVFilterContext* filter);

static if(FF_API_AVFILTERBUFFER){
    /**
     * Create a buffer reference wrapped around an already allocated image
     * buffer.*
     
     * @param data pointers to the planes of the image to reference
     * @param linesize linesizes for the planes of the image to reference
     * @param perms the required access permissions
     * @param w the width of the image specified by the data and linesize arrays
     * @param h the height of the image specified by the data and linesize arrays
     * @param format the pixel format of the image specified by the data and linesize arrays
     */
    AVFilterBufferRef* 
    avfilter_get_video_buffer_ref_from_arrays(ubyte*[4] data, int[4] linesize, int perms,
                                              int w, int h, AVPixelFormat format);

    /**
     * Create an audio buffer reference wrapped around an already
     * allocated samples buffer.*
     
     * @param data           pointers to the samples plane buffers
     * @param linesize       linesize for the samples plane buffers
     * @param perms          the required access permissions
     * @param nb_samples     number of samples per channel
     * @param sample_fmt     the format of each sample in the buffer to allocate
     * @param channel_layout the channel layout of the buffer
     */
    AVFilterBufferRef* avfilter_get_audio_buffer_ref_from_arrays(ubyte** data,
                                                                 int linesize,
                                                                 int perms,
                                                                 int nb_samples,
                                                                 AVSampleFormat sample_fmt,
                                                                 ulong channel_layout);
}

/** Initialize the filter system. Register all builtin filters. */
void avfilter_register_all();

static if(FF_API_OLD_FILTER_REGISTER){
    /** Uninitialize the filter system. Unregister all filters. */
    void avfilter_uninit();
}
/**
 * Register a filter. This is only needed if you plan to use
 * avfilter_get_by_name later to lookup the AVFilter structure by name. A
 * filter can still by instantiated with avfilter_graph_alloc_filter even if it
 * is not registered.*
 
 * @param filter the filter to register
 * @return 0 if the registration was succesfull, a negative value
 * otherwise
 */
int avfilter_register(AVFilter* filter);

/**
 * Get a filter definition matching the given name.*
 
 * @param name the filter name to find
 * @return     the filter definition, if any matching one is registered.
 *             NULL if none found.
 */
AVFilter* avfilter_get_by_name(char* name);

/**
 * Iterate over all registered filters.
 * @return If prev is non-NULL, next registered filter after prev or NULL if
 * prev is the last filter. If prev is NULL, return the first registered filter.
 */
AVFilter* avfilter_next(AVFilter* prev);

static if(FF_API_OLD_FILTER_REGISTER){
    /**
     * If filter is NULL, returns a pointer to the first registered filter pointer,
     * if filter is non-NULL, returns the next pointer after filter.
     * If the returned pointer points to NULL, the last registered filter
     * was already reached.
     * @deprecated use avfilter_next()
     */
    AVFilter** av_filter_next(AVFilter** filter);
}

static if(FF_API_AVFILTER_OPEN){
    /**
     * Create a filter instance.*
     
     * @param filter_ctx put here a pointer to the created filter context
     * on success, NULL on failure
     * @param filter    the filter to create an instance of
     * @param inst_name Name to give to the new instance. Can be NULL for none.
     * @return >= 0 in case of success, a negative error code otherwise
     * @deprecated use avfilter_graph_alloc_filter() instead
     */
    int avfilter_open(AVFilterContext** filter_ctx, AVFilter* filter, char* inst_name);
}

static if(FF_API_AVFILTER_INIT_FILTER){
    /**
     * Initialize a filter.*
     
     * @param filter the filter to initialize
     * @param args   A string of parameters to use when initializing the filter.
     *               The format and meaning of this string varies by filter.
     * @param opaque Any extra non-string data needed by the filter. The meaning
     *               of this parameter varies by filter.
     * @return       zero on success
     */
    int avfilter_init_filter(AVFilterContext* filter, char* args, void* opaque);
}
/**
 * Initialize a filter with the supplied parameters.*
 
 * @param ctx  uninitialized filter context to initialize
 * @param args Options to initialize the filter with. This must be a
 *             ':'-separated list of options in the 'key=value' form.
 *             May be NULL if the options have been set directly using the
 *             AVOptions API or there are no options that need to be set.
 * @return 0 on success, a negative AVERROR on failure
 */
int avfilter_init_str(AVFilterContext* ctx, char* args);

/**
 * Initialize a filter with the supplied dictionary of options.*
 
 * @param ctx     uninitialized filter context to initialize
 * @param options An AVDictionary filled with options for this filter. On
 *                return this parameter will be destroyed and replaced with
 *                a dict containing options that were not found. This dictionary
 *                must be freed by the caller.
 *                May be NULL, then this function is equivalent to
 *                avfilter_init_str() with the second parameter set to NULL.
 * @return 0 on success, a negative AVERROR on failure*
 
 * @note This function and avfilter_init_str() do essentially the same thing,
 * the difference is in manner in which the options are passed. It is up to the
 * calling code to choose whichever is more preferable. The two functions also
 * behave differently when some of the provided options are not declared as
 * supported by the filter. In such a case, avfilter_init_str() will fail, but
 * this function will leave those extra options in the options AVDictionary and
 * continue as usual.
 */
int avfilter_init_dict(AVFilterContext* ctx, AVDictionary** options);

/**
 * Free a filter context. This will also remove the filter from its
 * filtergraph's list of filters.*
 
 * @param filter the filter to free
 */
void avfilter_free(AVFilterContext* filter);

/**
 * Insert a filter in the middle of an existing link.*
 
 * @param link the link into which the filter should be inserted
 * @param filt the filter to be inserted
 * @param filt_srcpad_idx the input pad on the filter to connect
 * @param filt_dstpad_idx the output pad on the filter to connect
 * @return     zero on success
 */
int avfilter_insert_filter(AVFilterLink* link, AVFilterContext* filt,
                           uint filt_srcpad_idx, uint filt_dstpad_idx);

static if(FF_API_AVFILTERBUFFER){
    /**
     * Copy the frame properties of src to dst, without copying the actual
     * image data.*
     
     * @return 0 on success, a negative number on error.
     */
    int avfilter_copy_frame_props(AVFilterBufferRef* dst, in AVFrame* src);

    /**
     * Copy the frame properties and data pointers of src to dst, without copying
     * the actual data.*
     
     * @return 0 on success, a negative number on error.
     */
    int avfilter_copy_buf_props(AVFrame* dst, in AVFilterBufferRef* src);
}
/**
 * @return AVClass for AVFilterContext.*
 
 * @see av_opt_find().
 */
AVClass* avfilter_get_class();

struct AVFilterGraphInternal{}

struct AVFilterGraph {
    AVClass* av_class;
    static if(FF_API_FOO_COUNT){
        uint filter_count;
        AVFilterContext** filters;
    }else{
        uint nb_filters;
    }
    char* scale_sws_opts; ///< sws options to use for the auto-inserted scale filters
    char* resample_lavr_opts;   ///< libavresample options to use for the auto-inserted resample filters
    static if(FF_API_FOO_COUNT){
        uint nb_filters;
    }
    /**
     * Type of multithreading allowed for filters in this graph. A combination
     * of AVFILTER_THREAD_* flags.*
     
     * May be set by the caller at any point, the setting will apply to all
     * filters initialized after that. The default is allowing everything.*
     
     * When a filter in this graph is initialized, this field is combined using
     * bit AND with AVFilterContext.thread_type to get the final mask used for
     * determining allowed threading types. I.e. a threading type needs to be
     * set in both to be allowed.
     */
    int thread_type;

    /**
     * Maximum number of threads used by filters in this graph. May be set by
     * the caller before adding any filters to the filtergraph. Zero (the
     * default) means that the number of threads is determined automatically.
     */
    int nb_threads;

    /**
     * Opaque object for libavfilter internal use.
     */
    AVFilterGraphInternal* internal;
}

/**
 * Allocate a filter graph.
 */
AVFilterGraph* avfilter_graph_alloc();

/**
 * Create a new filter instance in a filter graph.*
 
 * @param graph graph in which the new filter will be used
 * @param filter the filter to create an instance of
 * @param name Name to give to the new instance (will be copied to
 *             AVFilterContext.name). This may be used by the caller to identify
 *             different filters, libavfilter itself assigns no semantics to
 *             this parameter. May be NULL.*
 
 * @return the context of the newly created filter instance (note that it is
 *         also retrievable directly through AVFilterGraph.filters or with
 *         avfilter_graph_get_filter()) on success or NULL or failure.
 */
AVFilterContext* avfilter_graph_alloc_filter(AVFilterGraph* graph,
                                             AVFilter* filter,
                                             char* name);

/**
 * Get a filter instance with name name from graph.*
 
 * @return the pointer to the found filter instance or NULL if it
 * cannot be found.
 */
AVFilterContext* avfilter_graph_get_filter(AVFilterGraph* graph, char* name);

static if(FF_API_AVFILTER_OPEN){
/**
 * Add an existing filter instance to a filter graph.*
 
 * @param graphctx  the filter graph
 * @param filter the filter to be added*
 
 * @deprecated use avfilter_graph_alloc_filter() to allocate a filter in a
 * filter graph
 */
int avfilter_graph_add_filter(AVFilterGraph* graphctx, AVFilterContext* filter);
}

/**
 * Create and add a filter instance into an existing graph.
 * The filter instance is created from the filter filt and inited
 * with the parameters args and opaque.*
 
 * In case of success put in* filt_ctx the pointer to the created
 * filter instance, otherwise set* filt_ctx to NULL.*
 
 * @param name the instance name to give to the created filter instance
 * @param graph_ctx the filter graph
 * @return a negative AVERROR error code in case of failure, a non
 * negative value otherwise
 */
int avfilter_graph_create_filter(AVFilterContext** filt_ctx, AVFilter* filt,
                                 char* name, char* args, void* opaque,
                                 AVFilterGraph* graph_ctx);

/**
 * Check validity and configure all the links and formats in the graph.*
 
 * @param graphctx the filter graph
 * @param log_ctx context used for logging
 * @return 0 in case of success, a negative AVERROR code otherwise
 */
int avfilter_graph_config(AVFilterGraph* graphctx, void* log_ctx);

/**
 * Free a graph, destroy its links, and set* graph to NULL.
 * If* graph is NULL, do nothing.
 */
void avfilter_graph_free(AVFilterGraph** graph);

/**
 * A linked-list of the inputs/outputs of the filter chain.*
 
 * This is mainly useful for avfilter_graph_parse() / avfilter_graph_parse2(),
 * where it is used to communicate open (unlinked) inputs and outputs from and
 * to the caller.
 * This struct specifies, per each not connected pad contained in the graph, the
 * filter context and the pad index required for establishing a link.
 */
struct AVFilterInOut {
    /** unique name for this input/output in the list */
    char* name;

    /** filter context associated to this input/output */
    AVFilterContext* filter_ctx;

    /** index of the filt_ctx pad to use for linking */
    int pad_idx;

    /** next input/input in the list, NULL if this is the last */
    AVFilterInOut* next;
}

/**
 * Allocate a single AVFilterInOut entry.
 * Must be freed with avfilter_inout_free().
 * @return allocated AVFilterInOut on success, NULL on failure.
 */
AVFilterInOut* avfilter_inout_alloc();

/**
 * Free the supplied list of AVFilterInOut and set* inout to NULL.
 * If* inout is NULL, do nothing.
 */
void avfilter_inout_free(AVFilterInOut** _inout);

/**
 * Add a graph described by a string to a graph.*
 
 * @param graph   the filter graph where to link the parsed graph context
 * @param filters string to be parsed
 * @param inputs  linked list to the inputs of the graph
 * @param outputs linked list to the outputs of the graph
 * @return zero on success, a negative AVERROR code on error
 */
int avfilter_graph_parse(AVFilterGraph* graph, char* filters,
                         AVFilterInOut* inputs, AVFilterInOut* outputs,
                         void* log_ctx);

/**
 * Add a graph described by a string to a graph.*
 
 * @param[in]  graph   the filter graph where to link the parsed graph context
 * @param[in]  filters string to be parsed
 * @param[out] inputs  a linked list of all free (unlinked) inputs of the
 *                     parsed graph will be returned here. It is to be freed
 *                     by the caller using avfilter_inout_free().
 * @param[out] outputs a linked list of all free (unlinked) outputs of the
 *                     parsed graph will be returned here. It is to be freed by the
 *                     caller using avfilter_inout_free().
 * @return zero on success, a negative AVERROR code on error*
 
 * @note the difference between avfilter_graph_parse2() and
 * avfilter_graph_parse() is that in avfilter_graph_parse(), the caller provides
 * the lists of inputs and outputs, which therefore must be known before calling
 * the function. On the other hand, avfilter_graph_parse2() \em returns the
 * inputs and outputs that are left unlinked after parsing the graph and the
 * caller then deals with them. Another difference is that in
 * avfilter_graph_parse(), the inputs parameter describes inputs of the
 * <em>already existing</em> part of the graph; i.e. from the point of view of
 * the newly created part, they are outputs. Similarly the outputs parameter
 * describes outputs of the already existing filters, which are provided as
 * inputs to the parsed filters.
 * avfilter_graph_parse2() takes the opposite approach -- it makes no reference
 * whatsoever to already existing parts of the graph and the inputs parameter
 * will on return contain inputs of the newly parsed part of the graph.
 * Analogously the outputs parameter will contain outputs of the newly created
 * filters.
 */
int avfilter_graph_parse2(AVFilterGraph* graph, char* filters,
                          AVFilterInOut** inputs,
                          AVFilterInOut** outputs);

/**
 * @}
 */

/* AVFILTER_AVFILTER_H */
