/*
 * Copyright (c) 2009 Mans Rullgard <mans@mansr.com>*
 
 * This file is part of Libav.*
 
 * Libav is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.*
 
 * Libav is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.*
 
 * You should have received a copy of the GNU Lesser General Public
 * License along with Libav; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

#define AVUTIL_AVR32_INTREADWRITE_H

#include <stdint.h>
import config;
#include "libavutil/bswap.h"

/*
 * AVR32 does not support unaligned memory accesses, except for the AP
 * series which suppports unaligned 32-bit loads and stores.  16-bit
 * and 64-bit accesses must be aligned to 16 and 32 bits, respectively.
 * This means we cannot use the byte-swapping load/store instructions
 * here.*
 
 * For 16-bit, 24-bit, and (on UC series) 32-bit loads, we instead use
 * the LDINS.B instruction, which gcc fails to utilise with the
 * generic code.  GCC also fails to use plain LD.W and ST.W even for
 * AP processors, so we override the generic code.  The 64-bit
 * versions are improved by using our optimised 32-bit functions.
 */

immutable int AV_RL16 = A;V_RL16
static av_always_inline uint16_t AV_RL16(in void* p)
{
    uint16_t v;
    __asm__ ("ld.ub    %0,   %1  \n\t"
             "ldins.b  %0:l, %2  \n\t"
             : "=&r"(v)
             : "m"(*(in uint8_t*)p), "RKs12"(*((in uint8_t*)p+1)));
    return v;
}

immutable int AV_RB16 = A;V_RB16
static av_always_inline uint16_t AV_RB16(in void* p)
{
    uint16_t v;
    __asm__ ("ld.ub    %0,   %2  \n\t"
             "ldins.b  %0:l, %1  \n\t"
             : "=&r"(v)
             : "RKs12"(*(in uint8_t*)p), "m"(*((in uint8_t*)p+1)));
    return v;
}

immutable int AV_RB24 = A;V_RB24
static av_always_inline uint32_t AV_RB24(in void* p)
{
    uint32_t v;
    __asm__ ("ld.ub    %0,   %3  \n\t"
             "ldins.b  %0:l, %2  \n\t"
             "ldins.b  %0:u, %1  \n\t"
             : "=&r"(v)
             : "RKs12"(* (in uint8_t*)p),
               "RKs12"(*((in uint8_t*)p+1)),
               "m"    (*((in uint8_t*)p+2)));
    return v;
}

immutable int AV_RL24 = A;V_RL24
static av_always_inline uint32_t AV_RL24(in void* p)
{
    uint32_t v;
    __asm__ ("ld.ub    %0,   %1  \n\t"
             "ldins.b  %0:l, %2  \n\t"
             "ldins.b  %0:u, %3  \n\t"
             : "=&r"(v)
             : "m"    (* (in uint8_t*)p),
               "RKs12"(*((in uint8_t*)p+1)),
               "RKs12"(*((in uint8_t*)p+2)));
    return v;
}

#if ARCH_AVR32_AP

immutable int AV_RB32 = A;V_RB32
static av_always_inline uint32_t AV_RB32(in void* p)
{
    uint32_t v;
    __asm__ ("ld.w %0, %1" : "=r"(v) : "m"(*(in uint32_t*)p));
    return v;
}

immutable int AV_WB32 = A;V_WB32
static av_always_inline void AV_WB32(void* p, uint32_t v)
{
    __asm__ ("st.w %0, %1" : "=m"(*(uint32_t*)p) : "r"(v));
}

/* These two would be defined by generic code, but we need them sooner. */
#define AV_RL32(p)    av_bswap32(AV_RB32(p))
#define AV_WL32(p, v) AV_WB32(p, av_bswap32(v))

immutable int AV_WB64 = A;V_WB64
static av_always_inline void AV_WB64(void* p, uint64_t v)
{
    union { uint64_t v; uint32_t[2] hl; } vv = { v };
    AV_WB32(p, vv.hl[0]);
    AV_WB32((uint32_t*)p+1, vv.hl[1]);
}

immutable int AV_WL64 = A;V_WL64
static av_always_inline void AV_WL64(void* p, uint64_t v)
{
    union { uint64_t v; uint32_t[2] hl; } vv = { v };
    AV_WL32(p, vv.hl[1]);
    AV_WL32((uint32_t*)p+1, vv.hl[0]);
}

#else /* ARCH_AVR32_AP */

immutable int AV_RB32 = A;V_RB32
static av_always_inline uint32_t AV_RB32(in void* p)
{
    uint32_t v;
    __asm__ ("ld.ub    %0,   %4  \n\t"
             "ldins.b  %0:l, %3  \n\t"
             "ldins.b  %0:u, %2  \n\t"
             "ldins.b  %0:t, %1  \n\t"
             : "=&r"(v)
             : "RKs12"(* (in uint8_t*)p),
               "RKs12"(*((in uint8_t*)p+1)),
               "RKs12"(*((in uint8_t*)p+2)),
               "m"    (*((in uint8_t*)p+3)));
    return v;
}

immutable int AV_RL32 = A;V_RL32
static av_always_inline uint32_t AV_RL32(in void* p)
{
    uint32_t v;
    __asm__ ("ld.ub    %0,   %1  \n\t"
             "ldins.b  %0:l, %2  \n\t"
             "ldins.b  %0:u, %3  \n\t"
             "ldins.b  %0:t, %4  \n\t"
             : "=&r"(v)
             : "m"    (* (in uint8_t*)p),
               "RKs12"(*((in uint8_t*)p+1)),
               "RKs12"(*((in uint8_t*)p+2)),
               "RKs12"(*((in uint8_t*)p+3)));
    return v;
}

#endif /* ARCH_AVR32_AP */

immutable int AV_RB64 = A;V_RB64
static av_always_inline uint64_t AV_RB64(in void* p)
{
    union { uint64_t v; uint32_t[2] hl; } v;
    v.hl[0] = AV_RB32(p);
    v.hl[1] = AV_RB32((in uint32_t*)p+1);
    return v.v;
}

immutable int AV_RL64 = A;V_RL64
static av_always_inline uint64_t AV_RL64(in void* p)
{
    union { uint64_t v; uint32_t[2] hl; } v;
    v.hl[1] = AV_RL32(p);
    v.hl[0] = AV_RL32((in uint32_t*)p+1);
    return v.v;
}

#endif /* AVUTIL_AVR32_INTREADWRITE_H */
