module libav.libavutil.avutil;
/*
 * copyright (c) 2006 Michael Niedermayer <michaelni@gmx.at>*
 
 * This file is part of Libav.*
 
 * Libav is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.*
 
 * Libav is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.*
 
 * You should have received a copy of the GNU Lesser General Public
 * License along with Libav; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

/**
 * @file
 * external API header
 */

import std.conv;
import libav.libavutil.rational;

extern (C):

/**
 * @mainpage*
 
 * @section libav_intro Introduction*
 
 * This document describes the usage of the different libraries
 * provided by Libav.*
 
 * @li @ref libavc "libavcodec" encoding/decoding library
 * @li @ref lavfi "libavfilter" graph-based frame editing library
 * @li @ref libavf "libavformat" I/O and muxing/demuxing library
 * @li @ref lavd "libavdevice" special devices muxing/demuxing library
 * @li @ref lavu "libavutil" common utility library
 * @li @ref lavr "libavresample" audio resampling, format conversion and mixing
 * @li @subpage libswscale  color conversion and scaling library*
 
 * @section libav_versioning Versioning and compatibility*
 
 * Each of the Libav libraries contains a version.h header, which defines a
 * major, minor and micro version number with the
 * <em>LIBRARYNAME_VERSION_{MAJOR,MINOR,MICRO}</em> macros. The major version
 * number is incremented with backward incompatible changes - e.g. removing
 * parts of the public API, reordering public struct members, etc. The minor
 * version number is incremented for backward compatible API changes or major
 * new features - e.g. adding a new public function or a new decoder. The micro
 * version number is incremented for smaller changes that a calling program
 * might still want to check for - e.g. changing behavior in a previously
 * unspecified situation.*
 
 * Libav guarantees backward API and ABI compatibility for each library as long
 * as its major version number is unchanged. This means that no public symbols
 * will be removed or renamed. Types and names of the public struct members and
 * values of public macros and enums will remain the same (unless they were
 * explicitly declared as not part of the public API). Documented behavior will
 * not change.*
 
 * In other words, any correct program that works with a given Libav snapshot
 * should work just as well without any changes with any later snapshot with the
 * same major versions. This applies to both rebuilding the program against new
 * Libav versions or to replacing the dynamic Libav libraries that a program
 * links against.*
 
 * However, new public symbols may be added and new members may be appended to
 * public structs whose size is not part of public ABI (most public structs in
 * Libav). New macros and enum values may be added. Behavior in undocumented
 * situations may change slightly (and be documented). All those are accompanied
 * by an entry in doc/APIchanges and incrementing either the minor or micro
 * version number.
 */

/**
 * @defgroup lavu Common utility functions*
 
 * @brief
 * libavutil contains the code shared across all the other Libav
 * libraries*
 
 * @note In order to use the functions provided by avutil you must include
 * the specific header.*
 
 * @{*
 
 * @defgroup lavu_crypto Crypto and Hashing*
 
 * @{
 * @}*
 
 * @defgroup lavu_math Maths
 * @{*
 
 * @}*
 
 * @defgroup lavu_string String Manipulation*
 
 * @{*
 
 * @}*
 
 * @defgroup lavu_mem Memory Management*
 
 * @{*
 
 * @}*
 
 * @defgroup lavu_data Data Structures
 * @{*
 
 * @}*
 
 * @defgroup lavu_audio Audio related*
 
 * @{*
 
 * @}*
 
 * @defgroup lavu_error Error Codes*
 
 * @{*
 
 * @}*
 
 * @defgroup lavu_misc Other*
 
 * @{*
 
 * @defgroup lavu_internal Internal*
 
 * Not exported functions, for internal usage only*
 
 * @{*
 
 * @}
 */


/**
 * @defgroup preproc_misc Preprocessor String Macros*
 
 * String manipulation macros*
 
 * @{
 */

alias AV_TOSTRING AV_STRINGIFY;

string AV_TOSTRING(T)(T obj){
    return to!string(obj);
}

string AV_GLUE(C : char)(C[] a, C[] b){
    return a ~ b;
}

alias AV_GLUE AV_JOIN;

//#define AV_PRAGMA(s) _Pragma(#s)

/**
 * @}
 */

/**
 * @defgroup version_utils Library Version Macros*
 
 * Useful to check and match library version in order to maintain
 * backward compatibility.*
 
 * @{
 */

T AV_VERSION_INT(T : int)(T a, T b, T c){
    return (a << 16 | b << 8 | c);
}
string AV_VERSION_DOT(T : int)(T a, T b, T c){
    return  AV_TOSTRING(a) ~ "." ~   AV_TOSTRING(b) ~ "." ~  AV_TOSTRING(c);
}

alias AV_VERSION_DOT AV_VERSION;

/**
 * @}
 */

/**
 * @addtogroup lavu_ver
 * @{
 */

/**
 * Return the LIBAVUTIL_VERSION_INT constant.
 */
uint avutil_version();

/**
 * Return the libavutil build-time configuration.
 */
char* avutil_configuration();

/**
 * Return the libavutil license.
 */
char* avutil_license();

/**
 * @}
 */

/**
 * @addtogroup lavu_media Media Type
 * @brief Media Type
 */

enum AVMediaType {
    AVMEDIA_TYPE_UNKNOWN = -1,  ///< Usually treated as AVMEDIA_TYPE_DATA
    AVMEDIA_TYPE_VIDEO,
    AVMEDIA_TYPE_AUDIO,
    AVMEDIA_TYPE_DATA,          ///< Opaque data information usually continuous
    AVMEDIA_TYPE_SUBTITLE,
    AVMEDIA_TYPE_ATTACHMENT,    ///< Opaque data information usually sparse
    AVMEDIA_TYPE_NB
}

/**
 * @defgroup lavu_in Constants
 * @{*
 
 * @defgroup lavu_enc Encoding specific*
 
 * @note those definition should move to avcodec
 * @{
 */

immutable int FF_LAMBDA_SHIFT = 7;
immutable int FF_LAMBDA_SCALE =  (1 << FF_LAMBDA_SHIFT);
immutable int FF_QP2LAMBDA = 118; ///< factor to convert from H.263 QP to lambda
immutable int FF_LAMBDA_MAX = (256 * 128 - 1);

immutable int FF_QUALITY_SCALE = FF_LAMBDA_SCALE; //FIXME maybe remove

/**
 * @}
 * @defgroup lavu_time Timestamp specific*
 
 * Libav internal timebase and timestamp definitions*
 
 * @{
 */

/**
 * @brief Undefined timestamp value*
 
 * Usually reported by demuxer that work on containers that do not provide
 * either pts or dts.
 */

immutable ulong AV_NOPTS_VALUE        = 0x8000000000000000;

/**
 * Internal time base represented as integer
 */

immutable int AV_TIME_BASE            = 1000000;

/**
 * Internal time base represented as fractional value
 */

immutable AVRational AV_TIME_BASE_Q   = {1, AV_TIME_BASE};

/**
 * @}
 * @}
 * @defgroup lavu_picture Image related*
 
 * AVPicture types, pixel formats and basic image planes manipulation.*
 
 * @{
 */

enum AVPictureType {
    AV_PICTURE_TYPE_I = 1, ///< Intra
    AV_PICTURE_TYPE_P,     ///< Predicted
    AV_PICTURE_TYPE_B,     ///< Bi-dir predicted
    AV_PICTURE_TYPE_S,     ///< S(GMC)-VOP MPEG4
    AV_PICTURE_TYPE_SI,    ///< Switching Intra
    AV_PICTURE_TYPE_SP,    ///< Switching Predicted
    AV_PICTURE_TYPE_BI,    ///< BI type
}

/**
 * Return a single letter to describe the given picture type
 * pict_type.*
 
 * @param[in] pict_type the picture type @return a single character
 * representing the picture type, '?' if pict_type is unknown
 */
char av_get_picture_type_char(AVPictureType pict_type);

/**
 * @}
 */

import libav.libavutil.error;
import libav.libavutil._version;

/**
 * @}
 * @}
 */

/* AVUTIL_AVUTIL_H */
