module libav.libavutil.common;
/*
 * copyright (c) 2006 Michael Niedermayer <michaelni@gmx.at>*
 
 * This file is part of Libav.*
 
 * Libav is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.*
 
 * Libav is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.*
 
 * You should have received a copy of the GNU Lesser General Public
 * License along with Libav; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

/**
 * @file
 * common internal and external API header
 */

import libav.libavutil.attributes;
import libav.libavutil._version;
import libav.libavutil.avconfig;

extern (C):

T AV_NE(T)(T be, T le)
{
    static if(AV_HAVE_BIGENDIAN){
        return be;
    }else{
        return le;
    }
}

//rounded division & shift
T RSHIFT(T)(T a, T b) if(__traits(isIntegral, T))
{
    return (a > 0 ? (a + ((1 << b) >> 1)) >> b : (a + ((1 << b) >> 1) - 1) >> b);
}
/* assume b>0 */
T ROUNDED_DIV(T)(T a,T b) if(__traits(isIntegral, T))
{
    return (a > 0 ? a + (b >> 1) : a - (b >> 1) / b);
}
T FFABS(T)(T a)
{
    return (a >= 0 ? a : (-a));
}

T FFSIGN(T)(T a)
{
    return (a > 0 ? 1 : -1);
}

T FFMAX(T)(T a,T b) if(__traits(isIntegral, T))
{
    return (a > b ? a : b);
}

T FFMAX3(T)(T a,T b, T c) if(__traits(isIntegral, T))
{
    return FFMAX(FFMAX(a,b),c);
}
T FFMIN(T)(T a,T b) if(__traits(isIntegral, T))
{
    return (a > b ? b : a);
}

T FFMIN3(T)(T a,T b, T c) if(__traits(isIntegral, T))
{
    return FFMIN(FFMIN(a,b),c);
}

void FFSWAP(T)(ref T a,ref T b)
{
    do{T SWAP_tmp= b; b= a; a= SWAP_tmp;}while(0);
}
uint FF_ARRAY_ELEMS(T)(T[] a)
{
    return (a.sizeof / a[0].sizeof);
}

T FFALIGN(T)(T x, T a) if(__traits(isIntegral, T))
{
    return ((x + a - 1) & ~(a - 1));
}

/* misc math functions */

static if(FF_API_AV_REVERSE){
    shared ubyte[256] av_reverse;
}

static if(HAVE_AV_CONFIG_H){
    import libav.libavutil.config;
    import libav.libavutil.intmath;
}

/* Pull in unguarded fallback defines at the end of this file. */
import libav.libavutil.common;

int av_log2(uint v);

int av_log2_16bit(uint v);

/**
 * Clip a signed integer value into the amin-amax range.
 * @param a value to clip
 * @param amin minimum value of the clip range
 * @param amax maximum value of the clip range
 * @return clipped value
 */
int av_clip(int a, int amin, int amax)
{
    if      (a < amin) return amin;
    else if (a > amax) return amax;
    else               return a;
}

/**
 * Clip a signed integer value into the 0-255 range.
 * @param a value to clip
 * @return clipped value
 */
ubyte av_clip_uint8(int a)
{
    if (a & (~0xFF)) return cast(ubyte)(-a >> 31);
    else             return cast(ubyte)a;
}

/**
 * Clip a signed integer value into the -128,127 range.
 * @param a value to clip
 * @return clipped value
 */
byte av_clip_int8(int a)
{
    if ((a + 0x80) & ~0xFF) return cast(byte)(a >> 31) ^ 0x7F;
    else                    return cast(byte)a;
}

/**
 * Clip a signed integer value into the 0-65535 range.
 * @param a value to clip
 * @return clipped value
 */
ushort av_clip_uint16(int a)
{
    if (a & ~0xFFFF) return cast(ushort)(-a >> 31);
    else             return cast(ushort)a;
}

/**
 * Clip a signed integer value into the -32768,32767 range.
 * @param a value to clip
 * @return clipped value
 */
short av_clip_int16(int a)
{
    if ((a + 0x8000) & ~0xFFFF) return cast(short)(a >> 31) ^ 0x7FFF;
    else                        return cast(short)a;
}

/**
 * Clip a signed 64-bit integer value into the -2147483648,2147483647 range.
 * @param a value to clip
 * @return clipped value
 */
int av_clipl_int32(long a)
{
    if ((a + 0x80000000U) & ~0xFFFFFFFFUL) return cast(int)(a >> 63) ^ 0x7FFFFFFF;
    else                                         return cast(int)a;
}

/**
 * Clip a signed integer to an upower of two range.
 * @param  a value to clip
 * @param  p bit position to clip at
 * @return clipped value
 */
uint av_clip_uintp2(int a, int p)
{
    if (a & ~((1 << p) - 1)) return -a >> 31 & ((1 << p) - 1);
    else                     return  a;
}

/**
 * Add two signed 32-bit values with saturation.*
 
 * @param  a one value
 * @param  b another value
 * @return sum with signed saturation
 */
int av_sat_add32(int a, int b)
{
    return av_clipl_int32(cast(long)a + b);
}

/**
 * Add a doubled value to another value with saturation at both stages.*
 
 * @param  a first value
 * @param  b value doubled and added to a
 * @return sum with signed saturation
 */
int av_sat_dadd32(int a, int b)
{
    return av_sat_add32(a, av_sat_add32(b, b));
}

/**
 * Clip a float value into the amin-amax range.
 * @param a value to clip
 * @param amin minimum value of the clip range
 * @param amax maximum value of the clip range
 * @return clipped value
 */
float av_clipf(float a, float amin, float amax)
{
    if      (a < amin) return amin;
    else if (a > amax) return amax;
    else               return a;
}

/** Compute ceil(log2(x)).
 * @param x value used to compute ceil(log2(x))
 * @return computed ceiling of log2(x)
 */
int av_ceil_log2(int x)
{
    return av_log2((x - 1) << 1);
}

/**
 * Count number of bits set to one in x
 * @param x value to count bits of
 * @return the number of bits set to one in x
 */
int av_popcount(uint x)
{
    x -= (x >> 1) & 0x55555555;
    x = (x & 0x33333333) + ((x >> 2) & 0x33333333);
    x = (x + (x >> 4)) & 0x0F0F0F0F;
    x += x >> 8;
    return (x + (x >> 16)) & 0x3F;
}

/**
 * Count number of bits set to one in x
 * @param x value to count bits of
 * @return the number of bits set to one in x
 */
int av_popcount64(ulong x)
{
    return av_popcount(cast(uint)x) + av_popcount(cast(uint)(x >> 32));
}

uint MKTAG(T)(T a,T b,T c,T d) if(is(T == int) || is(T == uint))
{
    return (a | (b << 8) | (c << 16) | (cast(uint)d << 24));
}

uint MKBETAG(T)(T a,T b,T c,T d) if(is(T == int) || is(T == uint))
{
    return (d | (c << 8) | (b << 16) | (cast(uint)a << 24));
}

/**
 * Convert a UTF-8 character (up to 4 bytes) to its 32-bit UCS-4 encoded form.*
 
 * @param val      Output value, must be an lvalue of type uint.
 * @param GET_BYTE Expression reading one byte from the input.
 *                 Evaluated up to 7 times (4 for the currently
 *                 assigned Unicode range).  With a memory buffer
 *                 input, this could be* ptr++.
 * @param ERROR    Expression to be evaluated on invalid input,
 *                 typically a goto statement.
 */
 /*
GET_UTF8(val, GET_BYTE, ERROR)
{
    val= GET_BYTE;
    {
        uint top = (val & 128) >> 1;
        if ((val & 0xc0) == 0x80)
            ERROR
        while (val & top) {
            int tmp= GET_BYTE - 128;
            if(tmp>>6)
                ERROR
            val= (val<<6) + tmp;
            top <<= 5;
        }
        val &= (top << 1) - 1;
    }
}
*/

/**
 * Convert a UTF-16 character (2 or 4 bytes) to its 32-bit UCS-4 encoded form.*
 
 * @param val       Output value, must be an lvalue of type uint.
 * @param GET_16BIT Expression returning two bytes of UTF-16 data converted
 *                  to native byte order.  Evaluated one or two times.
 * @param ERROR     Expression to be evaluated on invalid input,
 *                  typically a goto statement.
 */
/*
GET_UTF16(val, GET_16BIT, ERROR)
{
    val = GET_16BIT;
    {
        uint hi = val - 0xD800;
        if (hi < 0x800) {
            val = GET_16BIT - 0xDC00;
            if (val > 0x3FFU || hi > 0x3FFU)
                ERROR
            val += (hi<<10) + 0x10000;
        }
    }
}
*/

/**
 * @def PUT_UTF8(val, tmp, PUT_BYTE)
 * Convert a 32-bit Unicode character to its UTF-8 encoded form (up to 4 bytes long).
 * @param val is an input-only argument and should be of type uint. It holds
 * a UCS-4 encoded Unicode character that is to be converted to UTF-8. If
 * val is given as a function it is executed only once.
 * @param tmp is a temporary variable and should be of type uint8_t. It
 * represents an intermediate value during conversion that is to be
 * output by PUT_BYTE.
 * @param PUT_BYTE writes the converted UTF-8 bytes to any proper destination.
 * It could be a function or a statement, and uses tmp as the input byte.
 * For example, PUT_BYTE could be "*output++ = tmp;" PUT_BYTE will be
 * executed up to 4 times for values in the valid UTF-8 range and up to
 * 7 times in the general case, depending on the length of the converted
 * Unicode character.
 */
/*
PUT_UTF8(val, tmp, PUT_BYTE)
{
    {
        int bytes, shift;
        uint in = val;
        if (in < 0x80) {
            tmp = in;
            PUT_BYTE
        } else {
            bytes = (av_log2(in) + 4) / 5;
            shift = (bytes - 1) * 6;
            tmp = (256 - (256 >> bytes)) | (in >> shift);
            PUT_BYTE
            while (shift >= 6) {
                shift -= 6;
                tmp = 0x80 | ((in >> shift) & 0x3f);
                PUT_BYTE
            }
        }
    }
}
*/

/**
 * @def PUT_UTF16(val, tmp, PUT_16BIT)
 * Convert a 32-bit Unicode character to its UTF-16 encoded form (2 or 4 bytes).
 * @param val is an input-only argument and should be of type uint. It holds
 * a UCS-4 encoded Unicode character that is to be converted to UTF-16. If
 * val is given as a function it is executed only once.
 * @param tmp is a temporary variable and should be of type uint16_t. It
 * represents an intermediate value during conversion that is to be
 * output by PUT_16BIT.
 * @param PUT_16BIT writes the converted UTF-16 data to any proper destination
 * in desired endianness. It could be a function or a statement, and uses tmp
 * as the input byte.  For example, PUT_BYTE could be "*output++ = tmp;"
 * PUT_BYTE will be executed 1 or 2 times depending on input character.
 */
/*
PUT_UTF16(val, tmp, PUT_16BIT)
{
    {
        uint in = val;
        if (in < 0x10000) {
            tmp = in;
            PUT_16BIT
        } else {
            tmp = 0xD800 | ((in - 0x10000) >> 10);
            PUT_16BIT
            tmp = 0xDC00 | ((in - 0x10000) & 0x3FF);
            PUT_16BIT
        }
    }
}
*/

import libav.libavutil.mem;

static if(HAVE_AV_CONFIG_H){
    import libav.libavutil.internal;
}
/* HAVE_AV_CONFIG_H */

/* AVUTIL_COMMON_H */

/*
 * The following definitions are outside the multiple inclusion guard
 * to ensure they are immediately available in intmath.h.
 */
