module libjpeg.jpeglib;
/*
 * jpeglib.h
 *
 * Copyright (C) 1991-1998, Thomas G. Lane.
 * Modified 2002-2012 by Guido Vollbeding.
 * This file is part of the Independent JPEG Group's software.
 * For conditions of distribution and use, see the accompanying README file.*
 
 * This file defines the application interface for the JPEG library.
 * Most applications using the library need only include this file,
 * and perhaps jerror.h if they want to know the exact error codes.
 */

/*
 * First we include the configuration files that record how this
 * installation of the JPEG library is set up.  jconfig.h can be
 * generated automatically for many systems.  jmorecfg.h contains
 * manual configuration options that most people need not worry about.
 */

/* in case jinclude.h already did */
import libjpeg.jmorecfg;    /* seldom changed options */
import core.stdc.stdio;

extern (C):

/* Version IDs for the JPEG library.
 * Might be useful for tests like "#if JPEG_LIB_VERSION >= 90".
 */

immutable int JPEG_LIB_VERSION        = 90;  /* Compatibility version 9.0 */
immutable int JPEG_LIB_VERSION_MAJOR  = 9;
immutable int JPEG_LIB_VERSION_MINOR  = 0;


/* Various constants determining the sizes of things.
 * All of these are specified by the JPEG standard, so don't change them
 * if you want to be compatible.
 */

immutable int DCTSIZE              = 8;  /* The basic DCT block is 8x8 coefficients */
immutable int DCTSIZE2            = 64;  /* DCTSIZE squared; # of elements in a block */
immutable int NUM_QUANT_TBLS      = 4;  /* Quantization tables are numbered 0..3 */
immutable int NUM_HUFF_TBLS       = 4;  /* Huffman tables are numbered 0..3 */
immutable int NUM_ARITH_TBLS      = 16;  /* Arith-coding tables are numbered 0..15 */
immutable int MAX_COMPS_IN_SCAN   = 4;  /* JPEG limit on # of components in one scan */
immutable int MAX_SAMP_FACTOR     = 4;  /* JPEG limit on sampling factors */
/* Unfortunately, some bozo at Adobe saw no reason to be bound by the standard;
 * the PostScript DCT filter can emit files with many more than 10 blocks/MCU.
 * If you happen to run across such a file, you can up D_MAX_BLOCKS_IN_MCU
 * to handle it.  We even let you do this from the jconfig.h file.  However,
 * we strongly discourage changing C_MAX_BLOCKS_IN_MCU; just because Adobe
 * sometimes emits noncompliant files doesn't mean you should too.
 */
immutable int C_MAX_BLOCKS_IN_MCU   = 10; /* compressor's limit on blocks per MCU */
immutable int D_MAX_BLOCKS_IN_MCU   = 10; /* decompressor's limit on blocks per MCU */


/* Data structures for images (arrays of samples and of DCT coefficients).
 * On 80x86 machines, the image arrays are too big for near pointers,
 * but the pointer arrays can fit in near memory.
 */

alias ubyte* JSAMPROW;            /* ptr to one image row of pixel samples. */
alias JSAMPROW* JSAMPARRAY;         /* ptr to some rows (a 2-D sample array) */
alias JSAMPARRAY* JSAMPIMAGE;        /* a 3-D sample array: top index is color */

alias short[DCTSIZE2] JBLOCK;        /* one block of coefficients */
alias JBLOCK*         JBLOCKROW;    /* pointer to one row of coefficient blocks */
alias JBLOCKROW*     JBLOCKARRAY;    /* a 2-D array of coefficient blocks */
alias JBLOCKARRAY*    JBLOCKIMAGE;  /* a 3-D array of coefficient blocks */

alias short* shortPTR;  /* useful in a couple of places */


/* Types for JPEG compression parameters and working tables. */


/* DCT coefficient quantization tables. */

struct JQUANT_TBL{
  /* This array gives the coefficient quantizers in natural array order
   * (not the zigzag order in which they are stored in a JPEG DQT marker).
   * CAUTION: IJG versions prior to v6a kept this array in zigzag order.
   */
  ushort[DCTSIZE2] quantval;  /* quantization step for each coefficient */
  /* This field is used only during compression.  It's initialized FALSE when
   * the table is created, and set TRUE when it's been output to the file.
   * You could suppress output of a table by setting this to TRUE.
   * (See jpeg_suppress_tables for an example.)
   */
  int sent_table;    /* TRUE when table has been output */
}


/* Huffman coding tables. */

struct JHUFF_TBL{
  /* These two fields directly represent the contents of a JPEG DHT marker */
  ubyte[17] bits;    /* bits[k] = # of symbols with codes of */
        /* length k bits; bits[0] is unused */
  ubyte[256] huffval;    /* The symbols, in order of incr code length */
  /* This field is used only during compression.  It's initialized FALSE when
   * the table is created, and set TRUE when it's been output to the file.
   * You could suppress output of a table by setting this to TRUE.
   * (See jpeg_suppress_tables for an example.)
   */
  int sent_table;    /* TRUE when table has been output */
}


/* Basic info about one component (color channel). */

struct jpeg_component_info{
  /* These values are fixed over the whole image. */
  /* For compression, they must be supplied by parameter setup; */
  /* for decompression, they are read from the SOF marker. */
  int component_id;    /* identifier for this component (0..255) */
  int component_index;    /* its index in SOF or cinfo->comp_info[] */
  int h_samp_factor;    /* horizontal sampling factor (1..4) */
  int v_samp_factor;    /* vertical sampling factor (1..4) */
  int quant_tbl_no;    /* quantization table selector (0..3) */
  /* These values may vary between scans. */
  /* For compression, they must be supplied by parameter setup; */
  /* for decompression, they are read from the SOS marker. */
  /* The decompressor output side may not use these variables. */
  int dc_tbl_no;    /* DC entropy table selector (0..3) */
  int ac_tbl_no;    /* AC entropy table selector (0..3) */
  
  /* Remaining fields should be treated as private by applications. */
  
  /* These values are computed during compression or decompression startup: */
  /* Component's size in DCT blocks.
   * Any dummy blocks added to complete an MCU are not counted; therefore
   * these values do not depend on whether a scan is interleaved or not.
   */
  uint width_in_blocks;
  uint height_in_blocks;
  /* Size of a DCT block in samples,
   * reflecting any scaling we choose to apply during the DCT step.
   * Values from 1 to 16 are supported.
   * Note that different components may receive different DCT scalings.
   */
  int DCT_h_scaled_size;
  int DCT_v_scaled_size;
  /* The downsampled dimensions are the component's actual, unpadded number
   * of samples at the main buffer (preprocessing/compression interface);
   * DCT scaling is included, so
   * downsampled_width = ceil(image_width * Hi/Hmax * DCT_h_scaled_size/DCTSIZE)
   * and similarly for height.
   */
  uint downsampled_width;   /* actual width in samples */
  uint downsampled_height; /* actual height in samples */
  /* This flag is used only for decompression.  In cases where some of the
   * components will be ignored (eg grayscale output from YCbCr image),
   * we can skip most computations for the unused components.
   */
  int component_needed;  /* do we need the value of this component? */

  /* These values are computed before starting a scan of the component. */
  /* The decompressor output side may not use these variables. */
  int MCU_width;    /* number of blocks per MCU, horizontally */
  int MCU_height;    /* number of blocks per MCU, vertically */
  int MCU_blocks;    /* MCU_width * MCU_height */
  int MCU_sample_width;  /* MCU width in samples: MCU_width * DCT_h_scaled_size */
  int last_col_width;    /* # of non-dummy blocks across in last MCU */
  int last_row_height;    /* # of non-dummy blocks down in last MCU */

  /* Saved quantization table for component; NULL if none yet saved.
   * See jdinput.c comments about the need for this information.
   * This field is currently used only for decompression.
   */
  JQUANT_TBL* quant_table;

  /* Private per-component storage for DCT or IDCT subsystem. */
  void * dct_table;
};


/* The script for encoding a multiple-scan file is an array of these: */

struct jpeg_scan_info{
  int comps_in_scan;    /* number of components encoded in this scan */
  int[MAX_COMPS_IN_SCAN] component_index; /* their SOF/comp_info[] indexes */
  int Ss, Se;      /* progressive JPEG spectral selection parms */
  int Ah, Al;      /* progressive JPEG successive approx. parms */
}

/* The decompressor can save APPn and COM markers in a list of these: */

struct jpeg_marker_struct {
  jpeg_marker_struct* next;  /* next in list, or NULL */
  ubyte marker;      /* marker code: JPEG_COM, or JPEG_APP0+n */
  uint original_length;  /* # bytes of data in the file */
  uint data_length;  /* # bytes of data saved at data[] */
  ubyte* data;    /* the data contained in the marker */
  /* the marker length word is not counted in data_length or original_length */
}

/* Known color spaces. */

enum J_COLOR_SPACE{
  JCS_UNKNOWN,    /* error/unspecified */
  JCS_GRAYSCALE,    /* monochrome */
  JCS_RGB,    /* red/green/blue */
  JCS_YCbCr,    /* Y/Cb/Cr (also known as YUV) */
  JCS_CMYK,    /* C/M/Y/K */
  JCS_YCCK    /* Y/Cb/Cr/K */
}

/* Supported color transforms. */

enum J_COLOR_TRANSFORM{
  JCT_NONE           = 0,
  JCT_SUBTRACT_GREEN = 1
}

/* DCT/IDCT algorithm options. */

enum J_DCT_METHOD{
  JDCT_ISLOW,    /* slow but accurate integer algorithm */
  JDCT_IFAST,    /* faster, less accurate integer method */
  JDCT_FLOAT    /* floating-point: accurate, fast on fast HW */
}

/* may be overridden in jconfig.h */
/* may be overridden in jconfig.h */

/* Dithering options for decompression. */

enum J_DITHER_MODE{
  JDITHER_NONE,    /* no dithering */
  JDITHER_ORDERED,  /* simple ordered dither */
  JDITHER_FS    /* Floyd-Steinberg error diffusion dither */
}


/* Common fields between JPEG compression and decompression master structs. */
struct jpeg_common_struct {
  jpeg_error_mgr* err;  /* Error handler module */
  jpeg_memory_mgr* mem;  /* Memory manager module */
  jpeg_progress_mgr* progress; /* Progress monitor, or NULL if none */
  void* client_data;    /* Available for use by application */
  int is_decompressor;  /* So common code can tell which is which */
  int global_state;    /* For checking call sequence validity */
}
/* Routines that are to be used by both halves of the library are declared
 * to receive a pointer to this structure.  There are no actual instances of
 * jpeg_common_struct, only of jpeg_compress_struct and jpeg_decompress_struct.
 */
  /* Fields common to both master struct types */
  /* Additional fields follow in an actual jpeg_compress_struct or
   * jpeg_decompress_struct.  All three structs must agree on these
   * initial fields!  (This would be a lot cleaner in C++.)
   */

/* Master record for a compression instance */

struct jpeg_compress_struct {
  jpeg_error_mgr* err;  /* Error handler module */
  jpeg_memory_mgr* mem;  /* Memory manager module */
  jpeg_progress_mgr* progress; /* Progress monitor, or NULL if none */
  void* client_data;    /* Available for use by application */
  int is_decompressor;  /* So common code can tell which is which */
  int global_state;    /* For checking call sequence validity */

  /* Destination for compressed data */
  jpeg_destination_mgr* dest;

  /* Description of source image --- these fields must be filled in by
   * outer application before starting compression.  in_color_space must
   * be correct before you can even call jpeg_set_defaults().
   */

  uint image_width;  /* input image width */
  uint image_height;  /* input image height */
  int input_components;    /* # of color components in input image */
  J_COLOR_SPACE in_color_space;  /* colorspace of input image */

  double input_gamma;    /* image gamma of input image */

  /* Compression parameters --- these fields must be set before calling
   * jpeg_start_compress().  We recommend calling jpeg_set_defaults() to
   * initialize everything to reasonable defaults, then changing anything
   * the application specifically wants to change.  That way you won't get
   * burnt when new parameters are added.  Also note that there are several
   * helper routines to simplify changing parameters.
   */

  uint scale_num, scale_denom; /* fraction by which to scale image */

  uint jpeg_width;  /* scaled JPEG image width */
  uint jpeg_height;  /* scaled JPEG image height */
  /* Dimensions of actual JPEG image that will be written to file,
   * derived from input dimensions by scaling factors above.
   * These fields are computed by jpeg_start_compress().
   * You can also use jpeg_calc_jpeg_dimensions() to determine these values
   * in advance of calling jpeg_start_compress().
   */

  int data_precision;    /* bits of precision in image data */

  int num_components;    /* # of color components in JPEG image */
  J_COLOR_SPACE jpeg_color_space; /* colorspace of JPEG image */

  jpeg_component_info* comp_info;
  /* comp_info[i] describes component that appears i'th in SOF */

  JQUANT_TBL* quant_tbl_ptrs[NUM_QUANT_TBLS];
  int q_scale_factor[NUM_QUANT_TBLS];
  /* ptrs to coefficient quantization tables, or NULL if not defined,
   * and corresponding scale factors (percentage, initialized 100).
   */

  JHUFF_TBL* dc_huff_tbl_ptrs[NUM_HUFF_TBLS];
  JHUFF_TBL* ac_huff_tbl_ptrs[NUM_HUFF_TBLS];
  /* ptrs to Huffman coding tables, or NULL if not defined */

  ubyte arith_dc_L[NUM_ARITH_TBLS]; /* L values for DC arith-coding tables */
  ubyte arith_dc_U[NUM_ARITH_TBLS]; /* U values for DC arith-coding tables */
  ubyte arith_ac_K[NUM_ARITH_TBLS]; /* Kx values for AC arith-coding tables */

  int num_scans;    /* # of entries in scan_info array */
  jpeg_scan_info* scan_info; /* script for multi-scan file, or NULL */
  /* The default value of scan_info is NULL, which causes a single-scan
   * sequential JPEG file to be emitted.  To create a multi-scan file,
   * set num_scans and scan_info to point to an array of scan definitions.
   */

  int raw_data_in;    /* TRUE=caller supplies downsampled data */
  int arith_code;    /* TRUE=arithmetic coding, FALSE=Huffman */
  int optimize_coding;  /* TRUE=optimize entropy encoding parms */
  int CCIR601_sampling;  /* TRUE=first samples are cosited */
  int do_fancy_downsampling; /* TRUE=apply fancy downsampling */
  int smoothing_factor;    /* 1..100, or 0 for no input smoothing */
  J_DCT_METHOD dct_method;  /* DCT algorithm selector */

  /* The restart interval can be specified in absolute MCUs by setting
   * restart_interval, or in MCU rows by setting restart_in_rows
   * (in which case the correct restart_interval will be figured
   * for each scan).
   */
  uint restart_interval; /* MCUs per restart, or 0 for no restart */
  int restart_in_rows;    /* if > 0, MCU rows per restart interval */

  /* Parameters controlling emission of special markers. */

  int write_JFIF_header;  /* should a JFIF marker be written? */
  ubyte JFIF_major_version;  /* What to write for the JFIF version number */
  ubyte JFIF_minor_version;
  /* These three values are not used by the JPEG code, merely copied */
  /* into the JFIF APP0 marker.  density_unit can be 0 for unknown, */
  /* 1 for dots/inch, or 2 for dots/cm.  Note that the pixel aspect */
  /* ratio is defined by X_density/Y_density even when density_unit=0. */
  ubyte density_unit;    /* JFIF code for pixel size units */
  ushort X_density;    /* Horizontal pixel density */
  ushort Y_density;    /* Vertical pixel density */
  int write_Adobe_marker;  /* should an Adobe marker be written? */

  J_COLOR_TRANSFORM color_transform;
  /* Color transform identifier, writes LSE marker if nonzero */

  /* State variable: index of next scanline to be written to
   * jpeg_write_scanlines().  Application may use this to control its
   * processing loop, e.g., "while (next_scanline < image_height)".
   */

  uint next_scanline;  /* 0 .. image_height-1  */

  /* Remaining fields are known throughout compressor, but generally
   * should not be touched by a surrounding application.
   */

  /*
   * These fields are computed during compression startup
   */
  int progressive_mode;  /* TRUE if scan script uses progressive mode */
  int max_h_samp_factor;  /* largest h_samp_factor */
  int max_v_samp_factor;  /* largest v_samp_factor */

  int min_DCT_h_scaled_size;  /* smallest DCT_h_scaled_size of any component */
  int min_DCT_v_scaled_size;  /* smallest DCT_v_scaled_size of any component */

  uint total_iMCU_rows;  /* # of iMCU rows to be input to coef ctlr */
  /* The coefficient controller receives data in units of MCU rows as defined
   * for fully interleaved scans (whether the JPEG file is interleaved or not).
   * There are v_samp_factor * DCTSIZE sample rows of each component in an
   * "iMCU" (interleaved MCU) row.
   */
  
  /*
   * These fields are valid during any one scan.
   * They describe the components and MCUs actually appearing in the scan.
   */
  int comps_in_scan;    /* # of JPEG components in this scan */
  jpeg_component_info * cur_comp_info[MAX_COMPS_IN_SCAN];
  /** cur_comp_info[i] describes component that appears i'th in SOS */
  
  uint MCUs_per_row;  /* # of MCUs across the image */
  uint MCU_rows_in_scan;  /* # of MCU rows in the image */
  
  int blocks_in_MCU;    /* # of DCT blocks per MCU */
  int MCU_membership[C_MAX_BLOCKS_IN_MCU];
  /* MCU_membership[i] is index in cur_comp_info of component owning */
  /* i'th block in an MCU */

  int Ss, Se, Ah, Al;    /* progressive JPEG parameters for scan */

  int block_size;    /* the basic DCT block size: 1..16 */
  int* natural_order;  /* natural-order position array */
  int lim_Se;      /* min( Se, DCTSIZE2-1 ) */

  /*
   * Links to compression subobjects (methods and private variables of modules)
   */
  jpeg_comp_master * master;
  jpeg_c_main_controller * main;
  jpeg_c_prep_controller * prep;
  jpeg_c_coef_controller * coef;
  jpeg_marker_writer * marker;
  jpeg_color_converter * cconvert;
  jpeg_downsampler * downsample;
  jpeg_forward_dct * fdct;
  jpeg_entropy_encoder * entropy;
  jpeg_scan_info * script_space; /* workspace for jpeg_simple_progression */
  int script_space_size;
}


/* Master record for a decompression instance */

struct jpeg_decompress_struct {
  jpeg_error_mgr* err;  /* Error handler module */
  jpeg_memory_mgr* mem;  /* Memory manager module */
  jpeg_progress_mgr* progress; /* Progress monitor, or NULL if none */
  void* client_data;    /* Available for use by application */
  int is_decompressor;  /* So common code can tell which is which */
  int global_state;    /* For checking call sequence validity */

  /* Source of compressed data */
  jpeg_source_mgr* src;

  /* Basic description of image --- filled in by jpeg_read_header(). */
  /* Application may inspect these values to decide how to process image. */

  uint image_width;  /* nominal image width (from SOF marker) */
  uint image_height;  /* nominal image height */
  int num_components;    /* # of color components in JPEG image */
  J_COLOR_SPACE jpeg_color_space; /* colorspace of JPEG image */

  /* Decompression processing parameters --- these fields must be set before
   * calling jpeg_start_decompress().  Note that jpeg_read_header() initializes
   * them to default values.
   */

  J_COLOR_SPACE out_color_space; /* colorspace for output */

  uint scale_num, scale_denom; /* fraction by which to scale image */

  double output_gamma;    /* image gamma wanted in output */

  int buffered_image;  /* TRUE=multiple output passes */
  int raw_data_out;    /* TRUE=downsampled data wanted */

  J_DCT_METHOD dct_method;  /* IDCT algorithm selector */
  int do_fancy_upsampling;  /* TRUE=apply fancy upsampling */
  int do_block_smoothing;  /* TRUE=apply interblock smoothing */

  int quantize_colors;  /* TRUE=colormapped output wanted */
  /* the following are ignored if not quantize_colors: */
  J_DITHER_MODE dither_mode;  /* type of color dithering to use */
  int two_pass_quantize;  /* TRUE=use two-pass color quantization */
  int desired_number_of_colors;  /* max # colors to use in created colormap */
  /* these are significant only in buffered-image mode: */
  int enable_1pass_quant;  /* enable future use of 1-pass quantizer */
  int enable_external_quant;/* enable future use of external colormap */
  int enable_2pass_quant;  /* enable future use of 2-pass quantizer */

  /* Description of actual output image that will be returned to application.
   * These fields are computed by jpeg_start_decompress().
   * You can also use jpeg_calc_output_dimensions() to determine these values
   * in advance of calling jpeg_start_decompress().
   */

  uint output_width;  /* scaled image width */
  uint output_height;  /* scaled image height */
  int out_color_components;  /* # of color components in out_color_space */
  int output_components;  /* # of color components returned */
  /* output_components is 1 (a colormap index) when quantizing colors;
   * otherwise it equals out_color_components.
   */
  int rec_outbuf_height;  /* min recommended height of scanline buffer */
  /* If the buffer passed to jpeg_read_scanlines() is less than this many rows
   * high, space and time will be wasted due to unnecessary data copying.
   * Usually rec_outbuf_height will be 1 or 2, at most 4.
   */

  /* When quantizing colors, the output colormap is described by these fields.
   * The application can supply a colormap by setting colormap non-NULL before
   * calling jpeg_start_decompress; otherwise a colormap is created during
   * jpeg_start_decompress or jpeg_start_output.
   * The map has out_color_components rows and actual_number_of_colors columns.
   */
  int actual_number_of_colors;  /* number of entries in use */
  JSAMPARRAY colormap;    /* The color map as a 2-D pixel array */

  /* State variables: these variables indicate the progress of decompression.
   * The application may examine these but must not modify them.
   */

  /* Row index of next scanline to be read from jpeg_read_scanlines().
   * Application may use this to control its processing loop, e.g.,
   * "while (output_scanline < output_height)".
   */
  uint output_scanline;  /* 0 .. output_height-1  */

  /* Current input scan number and number of iMCU rows completed in scan.
   * These indicate the progress of the decompressor input side.
   */
  int input_scan_number;  /* Number of SOS markers seen so far */
  uint input_iMCU_row;  /* Number of iMCU rows completed */

  /* The "output scan number" is the notional scan being displayed by the
   * output side.  The decompressor will not allow output scan/row number
   * to get ahead of input scan/row, but it can fall arbitrarily far behind.
   */
  int output_scan_number;  /* Nominal scan number being displayed */
  uint output_iMCU_row;  /* Number of iMCU rows read */

  /* Current progression status.  coef_bits[c][i] indicates the precision
   * with which component c's DCT coefficient i (in zigzag order) is known.
   * It is -1 when no data has yet been received, otherwise it is the point
   * transform (shift) value for the most recent scan of the coefficient
   * (thus, 0 at completion of the progression).
   * This pointer is NULL when reading a non-progressive file.
   */
  int[DCTSIZE2]* coef_bits;  /* -1 or current Al value for each coef */

  /* Internal JPEG parameters --- the application usually need not look at
   * these fields.  Note that the decompressor output side may not use
   * any parameters that can change between scans.
   */

  /* Quantization and Huffman tables are carried forward across input
   * datastreams when processing abbreviated JPEG datastreams.
   */

  JQUANT_TBL*[NUM_QUANT_TBLS] quant_tbl_ptrs;
  /* ptrs to coefficient quantization tables, or NULL if not defined */

  JHUFF_TBL*[NUM_HUFF_TBLS] dc_huff_tbl_ptrs;
  JHUFF_TBL*[NUM_HUFF_TBLS] ac_huff_tbl_ptrs;
  /* ptrs to Huffman coding tables, or NULL if not defined */

  /* These parameters are never carried across datastreams, since they
   * are given in SOF/SOS markers or defined to be reset by SOI.
   */

  int data_precision;    /* bits of precision in image data */

  jpeg_component_info* comp_info;
  /* comp_info[i] describes component that appears i'th in SOF */

  int is_baseline;    /* TRUE if Baseline SOF0 encountered */
  int progressive_mode;  /* TRUE if SOFn specifies progressive mode */
  int arith_code;    /* TRUE=arithmetic coding, FALSE=Huffman */

  ubyte[NUM_ARITH_TBLS] arith_dc_L; /* L values for DC arith-coding tables */
  ubyte[NUM_ARITH_TBLS] arith_dc_U; /* U values for DC arith-coding tables */
  ubyte[NUM_ARITH_TBLS] arith_ac_K; /* Kx values for AC arith-coding tables */

  uint restart_interval; /* MCUs per restart interval, or 0 for no restart */

  /* These fields record data obtained from optional markers recognized by
   * the JPEG library.
   */
  int saw_JFIF_marker;  /* TRUE iff a JFIF APP0 marker was found */
  /* Data copied from JFIF marker; only valid if saw_JFIF_marker is TRUE: */
  ubyte JFIF_major_version;  /* JFIF version number */
  ubyte JFIF_minor_version;
  ubyte density_unit;    /* JFIF code for pixel size units */
  ushort X_density;    /* Horizontal pixel density */
  ushort Y_density;    /* Vertical pixel density */
  int saw_Adobe_marker;  /* TRUE iff an Adobe APP14 marker was found */
  ubyte Adobe_transform;  /* Color transform code from Adobe marker */

  J_COLOR_TRANSFORM color_transform;
  /* Color transform identifier derived from LSE marker, otherwise zero */

  int CCIR601_sampling;  /* TRUE=first samples are cosited */

  /* Aside from the specific data retained from APPn markers known to the
   * library, the uninterpreted contents of any or all APPn and COM markers
   * can be saved in a list for examination by the application.
   */
  jpeg_marker_struct* marker_list; /* Head of list of saved markers */

  /* Remaining fields are known throughout decompressor, but generally
   * should not be touched by a surrounding application.
   */

  /*
   * These fields are computed during decompression startup
   */
  int max_h_samp_factor;  /* largest h_samp_factor */
  int max_v_samp_factor;  /* largest v_samp_factor */

  int min_DCT_h_scaled_size;  /* smallest DCT_h_scaled_size of any component */
  int min_DCT_v_scaled_size;  /* smallest DCT_v_scaled_size of any component */

  uint total_iMCU_rows;  /* # of iMCU rows in image */
  /* The coefficient controller's input and output progress is measured in
   * units of "iMCU" (interleaved MCU) rows.  These are the same as MCU rows
   * in fully interleaved JPEG scans, but are used whether the scan is
   * interleaved or not.  We define an iMCU row as v_samp_factor DCT block
   * rows of each component.  Therefore, the IDCT output contains
   * v_samp_factor*DCT_v_scaled_size sample rows of a component per iMCU row.
   */

  ubyte * sample_range_limit; /* table for fast range-limiting */

  /*
   * These fields are valid during any one scan.
   * They describe the components and MCUs actually appearing in the scan.
   * Note that the decompressor output side must not use these fields.
   */
  int comps_in_scan;    /* # of JPEG components in this scan */
  jpeg_component_info * cur_comp_info[MAX_COMPS_IN_SCAN];
  /** cur_comp_info[i] describes component that appears i'th in SOS */

  uint MCUs_per_row;  /* # of MCUs across the image */
  uint MCU_rows_in_scan;  /* # of MCU rows in the image */

  int blocks_in_MCU;    /* # of DCT blocks per MCU */
  int MCU_membership[D_MAX_BLOCKS_IN_MCU];
  /* MCU_membership[i] is index in cur_comp_info of component owning */
  /* i'th block in an MCU */

  int Ss, Se, Ah, Al;    /* progressive JPEG parameters for scan */

  /* These fields are derived from Se of first SOS marker.
   */
  int block_size;    /* the basic DCT block size: 1..16 */
  int* natural_order; /* natural-order position array for entropy decode */
  int lim_Se;      /* min( Se, DCTSIZE2-1 ) for entropy decode */

  /* This field is shared between entropy decoder and marker parser.
   * It is either zero or the code of a JPEG marker that has been
   * read from the data source, but has not yet been processed.
   */
  int unread_marker;

  /*
   * Links to decompression subobjects (methods, private variables of modules)
   */
  jpeg_decomp_master* master;
  jpeg_d_main_controller* main;
  jpeg_d_coef_controller* coef;
  jpeg_d_post_controller* post;
  jpeg_input_controller* inputctl;
  jpeg_marker_reader* marker;
  jpeg_entropy_decoder* entropy;
  jpeg_inverse_dct* idct;
  jpeg_upsampler* upsample;
  jpeg_color_deconverter* cconvert;
  jpeg_color_quantizer* cquantize;
}


/* "Object" declarations for JPEG modules that may be supplied or called
 * directly by the surrounding application.
 * As with all objects in the JPEG library, these structs only define the
 * publicly visible methods and state variables of a module.  Additional
 * private fields may exist after the public ones.
 */


/* Error handler object */

immutable int JMSG_LENGTH_MAX  = 200;  /* recommended size of format_message buffer */
immutable int JMSG_STR_PARM_MAX  = 80;

private union _msg_parm{
  int[8] i;
  char[JMSG_STR_PARM_MAX] s;
} 

struct jpeg_error_mgr {
  /* Error exit handler: does not return to caller */
  void function(jpeg_common_struct* cinfo) error_exit;
  /* Conditionally emit a trace or warning message */
  void function(jpeg_common_struct* cinfo, int msg_level) emit_message;
  /* Routine that actually outputs a trace or error message */
  void function(jpeg_common_struct* cinfo) output_message;
  /* Format a message string for the most recent JPEG error or message */
  void function(jpeg_common_struct* cinfo, char * buffer) format_message;
  /* Reset error state variables at start of a new image */
  void function(jpeg_common_struct* cinfo) reset_error_mgr;
  
  /* The message ID code and any parameters are saved here.
   * A message can have one string parameter or up to 8 int parameters.
   */
  _msg_parm msg_parm;
  int msg_code;
  
  /* Standard state variables for error facility */
  
  int trace_level;    /* max msg_level that will be displayed */
  
  /* For recoverable corrupt-data errors, we emit a warning message,
   * but keep going unless emit_message chooses to abort.  emit_message
   * should count warnings in num_warnings.  The surrounding application
   * can check for bad data by seeing if num_warnings is nonzero at the
   * end of processing.
   */
  long num_warnings;    /* number of corrupt-data warnings */

  /* These fields point to the table(s) of error message strings.
   * An application can change the table pointer to switch to a different
   * message list (typically, to change the language in which errors are
   * reported).  Some applications may wish to add additional error codes
   * that will be handled by the JPEG library error mechanism; the second
   * table pointer is used for this purpose.*
   
   * First table includes all errors generated by JPEG library itself.
   * Error code 0 is reserved for a "no such error string" message.
   */
  char ** jpeg_message_table; /* Library errors */
  int last_jpeg_message;    /* Table contains strings 0..last_jpeg_message */
  /* Second table can be added by application (see cjpeg/djpeg for example).
   * It contains strings numbered first_addon_message..last_addon_message.
   */
  char ** addon_message_table; /* Non-library errors */
  int first_addon_message;  /* code for first string in addon table */
  int last_addon_message;  /* code for last string in addon table */
};


/* Progress monitor object */

struct jpeg_progress_mgr {
  void function(jpeg_common_struct* cinfo) progress_monitor;

  long pass_counter;    /* work units completed in this pass */
  long pass_limit;    /* total number of work units in this pass */
  int completed_passes;    /* passes completed so far */
  int total_passes;    /* total number of passes expected */
};


/* Data destination object for compression */

struct jpeg_destination_mgr {
  ubyte * next_output_byte;  /* => next byte to write in buffer */
  size_t free_in_buffer;  /* # of byte spaces remaining in buffer */

  void function(jpeg_compress_struct* cinfo) init_destination;
  int function(jpeg_compress_struct* cinfo) empty_output_buffer;
  void function(jpeg_compress_struct* cinfo) term_destination;
};


/* Data source object for decompression */

struct jpeg_source_mgr {
  ubyte* next_input_byte; /* => next byte to read from buffer */
  size_t bytes_in_buffer;  /* # of bytes remaining in buffer */

  void function(jpeg_decompress_struct* cinfo) init_source;
  int function(jpeg_decompress_struct* cinfo) fill_input_buffer;
  void function(jpeg_decompress_struct* cinfo, int num_bytes) skip_input_data;
  int function(jpeg_decompress_struct* cinfo, int desired) resync_to_restart;
  void function(jpeg_decompress_struct* cinfo) term_source;
}


/* Memory manager object.
 * Allocates "small" objects (a few K total), "large" objects (tens of K),
 * and "really big" objects (virtual arrays with backing store if needed).
 * The memory manager does not allow individual objects to be freed; rather,
 * each created object is assigned to a pool, and whole pools can be freed
 * at once.  This is faster and more convenient than remembering exactly what
 * to free, especially where malloc()/free() are not too speedy.
 * NB: alloc routines never return NULL.  They exit to error_exit if not
 * successful.
 */

immutable int JPOOL_PERMANENT  = 0;  /* lasts until master record is destroyed */
immutable int JPOOL_IMAGE  = 1;  /* lasts until done with image/datastream */
immutable int JPOOL_NUMPOOLS  = 2;

struct jpeg_memory_mgr {
  /* Method pointers */
  void* function(jpeg_common_struct* cinfo, int pool_id,
        size_t sizeofobject) alloc_small;
  void* function(jpeg_common_struct* cinfo, int pool_id,
             size_t sizeofobject) alloc_large;
  JSAMPARRAY function(jpeg_common_struct* cinfo, int pool_id,
             uint samplesperrow,
             uint numrows) alloc_sarray;
  JBLOCKARRAY function(jpeg_common_struct* cinfo, int pool_id,
              uint blocksperrow,
              uint numrows) alloc_barray;
  jvirt_sarray_control* function(jpeg_common_struct* cinfo,
              int pool_id,
              int pre_zero,
              uint samplesperrow,
              uint numrows,
              uint maxaccess) request_virt_sarray;
  jvirt_barray_control* function(jpeg_common_struct* cinfo,
              int pool_id,
              int pre_zero,
              uint blocksperrow,
              uint numrows,
              uint maxaccess) request_virt_barray;
  void function(jpeg_common_struct* cinfo) realize_virt_arrays;
  JSAMPARRAY function(jpeg_common_struct* cinfo,
             jvirt_sarray_control* ptr,
             uint start_row,
             uint num_rows,
             int writable) access_virt_sarray;
  JBLOCKARRAY function(jpeg_common_struct* cinfo,
              jvirt_barray_control* ptr,
              uint start_row,
              uint num_rows,
              int writable) access_virt_barray;
  void function(jpeg_common_struct* cinfo, int pool_id) free_pool;
  void function(jpeg_common_struct* cinfo) self_destruct;

  /* Limit on memory allocation for this JPEG object.  (Note that this is
   * merely advisory, not a guaranteed maximum; it only affects the space
   * used for virtual-array buffers.)  May be changed by outer application
   * after creating the JPEG object.
   */
  long max_memory_to_use;

  /* Maximum allocation request accepted by alloc_large. */
  long max_alloc_chunk;
}


/* Routine signature for application-supplied marker processing methods.
 * Need not pass marker code since it is stored in cinfo->unread_marker.
 */
alias int function(jpeg_decompress_struct* cinfo) jpeg_marker_parser_method;


/* Declarations for routines called by application.
 * The JPP macro hides prototype parameters from compilers that can't cope.
 * Note JPP requires double parentheses.
 */

/* Short forms of external names for systems with brain-damaged linkers.
 * We shorten external names to be unique in the first six letters, which
 * is good enough for all known systems.
 * (If your compiler itself needs names to be unique in less than 15 
 * characters, you are out of luck.  Get a better compiler.)
 */

alias jpeg_std_error jStdError;
alias jpeg_CreateCompress jCreaCompress;
alias jpeg_CreateDecompress jCreaDecompress;
alias jpeg_destroy_compress jDestCompress;
alias jpeg_destroy_decompress jDestDecompress;
alias jpeg_stdio_dest jStdDest;
alias jpeg_stdio_src jStdSrc;
alias jpeg_mem_dest jMemDest;
alias jpeg_mem_src jMemSrc;
alias jpeg_set_defaults jSetDefaults;
alias jpeg_set_colorspace jSetColorspace;
alias jpeg_default_colorspace jDefColorspace;
alias jpeg_set_quality jSetQuality;
alias jpeg_set_linear_quality jSetLQuality;
alias jpeg_default_qtables jDefQTables;
alias jpeg_add_quant_table jAddQuantTable;
alias jpeg_quality_scaling jQualityScaling;
alias jpeg_simple_progression jSimProgress;
alias jpeg_suppress_tables jSuppressTables;
alias jpeg_alloc_quant_table jAlcQTable;
alias jpeg_alloc_huff_table jAlcHTable;
alias jpeg_start_compress jStrtCompress;
alias jpeg_write_scanlines jWrtScanlines;
alias jpeg_finish_compress jFinCompress;
alias jpeg_calc_jpeg_dimensions jCjpegDimensions;
alias jpeg_write_raw_data jWrtRawData;
alias jpeg_write_marker jWrtMarker;
alias jpeg_write_m_header jWrtMHeader;
alias jpeg_write_m_byte jWrtMByte;
alias jpeg_write_tables jWrtTables;
alias jpeg_read_header jReadHeader;
alias jpeg_start_decompress jStrtDecompress;
alias jpeg_read_scanlines jReadScanlines;
alias jpeg_finish_decompress jFinDecompress;
alias jpeg_read_raw_data jReadRawData;
alias jpeg_has_multiple_scans jHasMultScn;
alias jpeg_start_output jStrtOutput;
alias jpeg_finish_output jFinOutput;
alias jpeg_input_complete jInComplete;
alias jpeg_new_colormap jNewCMap;
alias jpeg_consume_input jConsumeInput;
alias jpeg_core_output_dimensions jCoreDimensions;
alias jpeg_calc_output_dimensions jCalcDimensions;
alias jpeg_save_markers jSaveMarkers;
alias jpeg_set_marker_processor jSetMarker;
alias jpeg_read_coefficients jReadCoefs;
alias jpeg_write_coefficients jWrtCoefs;
alias jpeg_copy_critical_parameters jCopyCrit;
alias jpeg_abort_compress jAbrtCompress;
alias jpeg_abort_decompress jAbrtDecompress;
alias jpeg_abort jAbort;
alias jpeg_destroy jDestroy;
alias jpeg_resync_to_restart jResyncRestart;
/* NEED_SHORT_EXTERNAL_NAMES */


/* Default error-management setup */
jpeg_error_mgr* jpeg_std_error(jpeg_error_mgr* err);

/* Initialization of JPEG compression objects.
 * jpeg_create_compress() and jpeg_create_decompress() are the exported
 * names that applications should call.  These expand to calls on
 * jpeg_CreateCompress and jpeg_CreateDecompress with additional information
 * passed for version mismatch checking.
 * NB: you must set up the error-manager BEFORE calling jpeg_create_xxx.
 */
void jpeg_create_compress(jpeg_compress_struct* cinfo){
  jpeg_CreateCompress(cinfo, JPEG_LIB_VERSION,
                      cast(size_t)jpeg_compress_struct.sizeof);
}
void jpeg_create_decompress(jpeg_decompress_struct* cinfo){
  jpeg_CreateDecompress(cinfo, JPEG_LIB_VERSION,
                        cast(size_t)jpeg_compress_struct.sizeof);
}
void jpeg_CreateCompress (jpeg_compress_struct* cinfo,
              int _version, size_t structsize);
void jpeg_CreateDecompress (jpeg_decompress_struct* cinfo,
          int _version, size_t structsize);
/* Destruction of JPEG compression objects */
void jpeg_destroy_compress (jpeg_compress_struct* cinfo);
void jpeg_destroy_decompress (jpeg_decompress_struct* cinfo);

/* Standard data source and destination managers: stdio streams. */
/* Caller is responsible for opening the file before and closing after. */
void jpeg_stdio_dest (jpeg_compress_struct* cinfo, FILE* outfile);
void jpeg_stdio_src (jpeg_decompress_struct* cinfo, FILE* infile);

/* Data source and destination managers: memory buffers. */
void jpeg_mem_dest (jpeg_compress_struct* cinfo,
             ubyte* * outbuffer,
             ulong * outsize);
void jpeg_mem_src (jpeg_decompress_struct* cinfo,
            ubyte * inbuffer,
            ulong insize);

/* Default parameter setup for compression */
void jpeg_set_defaults (jpeg_compress_struct* cinfo);
/* Compression parameter setup aids */
void jpeg_set_colorspace (jpeg_compress_struct* cinfo,
              J_COLOR_SPACE colorspace);
void jpeg_default_colorspace (jpeg_compress_struct* cinfo);
void jpeg_set_quality (jpeg_compress_struct* cinfo, int quality,
           int force_baseline);
void jpeg_set_linear_quality (jpeg_compress_struct* cinfo,
            int scale_factor,
            int force_baseline);
void jpeg_default_qtables (jpeg_compress_struct* cinfo,
               int force_baseline);
void jpeg_add_quant_table (jpeg_compress_struct* cinfo, int which_tbl,
               in uint* basic_table,
               int scale_factor,
               int force_baseline);
int jpeg_quality_scaling (int quality);
void jpeg_simple_progression (jpeg_compress_struct* cinfo);
void jpeg_suppress_tables (jpeg_compress_struct* cinfo,
               int suppress);
JQUANT_TBL* jpeg_alloc_quant_table (jpeg_common_struct* cinfo);
JHUFF_TBL* jpeg_alloc_huff_table (jpeg_common_struct* cinfo);

/* Main entry points for compression */
void jpeg_start_compress (jpeg_compress_struct* cinfo,
              int write_all_tables);
uint jpeg_write_scanlines (jpeg_compress_struct* cinfo,
               JSAMPARRAY scanlines,
               uint num_lines);
void jpeg_finish_compress (jpeg_compress_struct* cinfo);

/* Precalculate JPEG dimensions for current compression parameters. */
void jpeg_calc_jpeg_dimensions (jpeg_compress_struct* cinfo);

/* Replaces jpeg_write_scanlines when writing raw downsampled data. */
uint jpeg_write_raw_data (jpeg_compress_struct* cinfo,
              JSAMPIMAGE data,
              uint num_lines);

/* Write a special marker.  See libjpeg.txt concerning safe usage. */
void jpeg_write_marker
  (jpeg_compress_struct* cinfo, int marker,
       in ubyte * dataptr, uint datalen);
/* Same, but piecemeal. */
void jpeg_write_m_header
  (jpeg_compress_struct* cinfo, int marker, uint datalen);
void jpeg_write_m_byte
  (jpeg_compress_struct* cinfo, int val);

/* Alternate compression function: just write an abbreviated table file */
void jpeg_write_tables (jpeg_compress_struct* cinfo);

/* Decompression startup: read start of JPEG datastream to see what's there */
int jpeg_read_header (jpeg_decompress_struct* cinfo,
          int require_image);
/* Return value is one of: */
immutable int JPEG_SUSPENDED    = 0; /* Suspended due to lack of input data */
immutable int JPEG_HEADER_OK    = 1; /* Found valid image datastream */
immutable int JPEG_HEADER_TABLES_ONLY  = 2; /* Found valid table-specs-only datastream */
/* If you pass require_image = TRUE (normal case), you need not check for
 * a TABLES_ONLY return code; an abbreviated file will cause an error exit.
 * JPEG_SUSPENDED is only possible if you use a data source module that can
 * give a suspension return (the stdio source module doesn't).
 */

/* Main entry points for decompression */
int jpeg_start_decompress (jpeg_decompress_struct* cinfo);
uint jpeg_read_scanlines (jpeg_decompress_struct* cinfo,
              JSAMPARRAY scanlines,
              uint max_lines);
int jpeg_finish_decompress (jpeg_decompress_struct* cinfo);

/* Replaces jpeg_read_scanlines when reading raw downsampled data. */
uint jpeg_read_raw_data (jpeg_decompress_struct* cinfo,
             JSAMPIMAGE data,
             uint max_lines);

/* Additional entry points for buffered-image mode. */
int jpeg_has_multiple_scans (jpeg_decompress_struct* cinfo);
int jpeg_start_output (jpeg_decompress_struct* cinfo,
               int scan_number);
int jpeg_finish_output (jpeg_decompress_struct* cinfo);
int jpeg_input_complete (jpeg_decompress_struct* cinfo);
void jpeg_new_colormap (jpeg_decompress_struct* cinfo);
int jpeg_consume_input (jpeg_decompress_struct* cinfo);
/* Return value is one of: */
/* immutable int JPEG_SUSPENDED  = 0;    Suspended due to lack of input data */
immutable int JPEG_REACHED_SOS  = 1; /* Reached start of new scan */
immutable int JPEG_REACHED_EOI  = 2; /* Reached end of image */
immutable int JPEG_ROW_COMPLETED  = 3; /* Completed one iMCU row */
immutable int JPEG_SCAN_COMPLETED  = 4; /* Completed last iMCU row of a scan */

/* Precalculate output dimensions for current decompression parameters. */
void jpeg_core_output_dimensions (jpeg_decompress_struct* cinfo);
void jpeg_calc_output_dimensions (jpeg_decompress_struct* cinfo);

/* Control saving of COM and APPn markers into marker_list. */
void jpeg_save_markers
  (jpeg_decompress_struct* cinfo, int marker_code,
       uint length_limit);

/* Install a special processing method for COM or APPn markers. */
void jpeg_set_marker_processor
  (jpeg_decompress_struct* cinfo, int marker_code,
       jpeg_marker_parser_method routine);

/* Read or write raw DCT coefficients --- useful for lossless transcoding. */
jvirt_barray_control** jpeg_read_coefficients (jpeg_decompress_struct* cinfo);
void jpeg_write_coefficients (jpeg_compress_struct* cinfo,
            jvirt_barray_control* * coef_arrays);
void jpeg_copy_critical_parameters (jpeg_decompress_struct* srcinfo,
            jpeg_compress_struct* dstinfo);

/* If you choose to abort compression or decompression before completing
 * jpeg_finish_(de)compress, then you need to clean up to release memory,
 * temporary files, etc.  You can just call jpeg_destroy_(de)compress
 * if you're done with the JPEG object, but if you want to clean it up and
 * reuse it, call this:
 */
void jpeg_abort_compress (jpeg_compress_struct* cinfo);
void jpeg_abort_decompress (jpeg_decompress_struct* cinfo);

/* Generic versions of jpeg_abort and jpeg_destroy that work on either
 * flavor of JPEG object.  These may be more convenient in some places.
 */
void jpeg_abort (jpeg_common_struct* cinfo);
void jpeg_destroy (jpeg_common_struct* cinfo);

/* Default restart-marker-resync procedure for use by data source modules */
int jpeg_resync_to_restart (jpeg_decompress_struct* cinfo,
              int desired);


/* These marker codes are exported since applications and data source modules
 * are likely to want to use them.
 */

immutable int JPEG_RST0  = 0xD0;  /* RST0 marker code */
immutable int JPEG_EOI  = 0xD9;  /* EOI marker code */
immutable int JPEG_APP0  = 0xE0;  /* APP0 marker code */
immutable int JPEG_COM  = 0xFE;  /* COM marker code */


/* If we have a brain-damaged compiler that emits warnings (or worse, errors)
 * for structure definitions that are never filled in, keep it quiet by
 * supplying dummy definitions for the various substructures.
 */

/* will be defined in jpegint.h */
struct jvirt_sarray_control { long dummy; };
struct jvirt_barray_control { long dummy; };
struct jpeg_comp_master { long dummy; };
struct jpeg_c_main_controller { long dummy; };
struct jpeg_c_prep_controller { long dummy; };
struct jpeg_c_coef_controller { long dummy; };
struct jpeg_marker_writer { long dummy; };
struct jpeg_color_converter { long dummy; };
struct jpeg_downsampler { long dummy; };
struct jpeg_forward_dct { long dummy; };
struct jpeg_entropy_encoder { long dummy; };
struct jpeg_decomp_master { long dummy; };
struct jpeg_d_main_controller { long dummy; };
struct jpeg_d_coef_controller { long dummy; };
struct jpeg_d_post_controller { long dummy; };
struct jpeg_input_controller { long dummy; };
struct jpeg_marker_reader { long dummy; };
struct jpeg_entropy_decoder { long dummy; };
struct jpeg_inverse_dct { long dummy; };
struct jpeg_upsampler { long dummy; };
struct jpeg_color_deconverter { long dummy; };
struct jpeg_color_quantizer { long dummy; };
/* JPEG_INTERNALS */
/* INCOMPLETE_TYPES_BROKEN */


/*
 * The JPEG library modules define JPEG_INTERNALS before including this file.
 * The internal structure declarations are read only when that is true.
 * Applications using the library should not include jpegint.h, but may wish
 * to include jerror.h.
 */


/* JPEGLIB_H */
