﻿/**
 * バイト列をファイルの読み書きと同じように扱うためのクラス
 * 
 * License:   Public Domain
 * Date:      August 21, 010
 * Author:    Kazuki Okawa, naka4959@kcn.ne.jp
 * 
 * Version:   0.01
 * History:
 *  v0.01 公開
 * 
*/
module ngbase.stringio;

import std.stream;
import std.c.string;

version(D_Version2){
  import core.memory;
}else{
  import std.GC;
}

debug{
  import std.stdio;
  alias std.stream.File File;
}

version(unittest){
  import std.c.stdio;
}

/** アクセスタイプ */
enum accessible : uint{
  READ = 1, /// 読み込み可能
  WRITE = 2, /// 書き込み可能
}

/**
 * メモリストリームのクラス
*/
class String_IO : Stream{
  private{
    ubyte[] data; // バッファの先頭位置
    ulong length; // バッファのサイズ
    ulong point; // 読み書きを行う位置
  }
  public{
    /**
     * 特定の領域を，String_IOとして宣言
     * 
     * Params:
     *  buffer = String_IOで管理するバッファ
     *  size = バッファの大きさ
     *  access = 読み書きが可能であるかどうか
     *
     * Throws:
     *  バッファの領域が存在しない場合，StreamException<br>
     *  読み書きのどちらも出来ない場合，StreamException
     *
     * Example: 
     *  外部で宣言した領域を，読み込みと読み書きのstringioとして宣言
     * ---
     * int size = 256;
     * ubyte* buffer = cast(ubyte*)malloc(size);
     * auto a = new String_IO(buffer, size, accessible.READ);
     * auto b = new String_IO(buffer, size, accessible.READ | accessible.WRITE);
     * ---
    */
    this(ubyte[] buffer, in accessible access){
      if(buffer.length == 0){
        throw(new StreamException("Buffer isn't exist"));
      }
      if((access & (accessible.READ | accessible.WRITE)) == false){
        throw(new StreamException("Can't use this object"));
      }
      if((access & accessible.READ) != 0){
        this.readable = true;
      }else{
        this.readable = false;
      }    
      if((access & accessible.WRITE) != 0){
        this.writeable = true;
      }else{
        this.writeable = false;
      }
      this.seekable = true;
      this.data = buffer;
      this.length = size;
      this.point = 0;
    }
    /**
     * 新しい領域を作り，String_IOとして宣言
     * Params:
     *  size = バッファの大きさ
     *  access = 読み書きが可能であるかどうか
     *
     * Throws:
     *  バッファの領域が存在しない場合，StreamException<br>
     *  読み書きのどちらも出来ない場合，StreamException
     *
     * Example: 
     *  書き込みと読み書きの二つのstringioを，新しい領域で宣言
     * ---
     * auto a = new String_IO(4096, accessible.WRITE);
     * auto b = new String_IO(4096, accessible.READ | accessible.WRITE);
     * ---
    */
    this(in uint size, in accessible access){
      this(new ubyte[size], access);
    }
    /**
     * バッファのポインタを取得する
     *
     * Returns:
     *  バッファの先頭位置
    */
    ubyte* buffer(){
      return(this.data.ptr);
    }
    /**
     * バッファからデータを読み込む<br>
     * 残り領域以上に読み込もうとした場合，残り領域全てを読み込む
     *
     * Params:
     *  buffer = データが読み込まれる領域の先頭ポインタ
     *  size = 読み込むデータの大きさ
     * 
     * Returns:
     *  読み込めたバイト数
     * 
     * Throws:
     *  読み込みを許可していない場合，ReadException
    */
    override size_t readBlock(void* buffer, size_t size){
      size_t result;
      assertReadable();
      if(size + this.point > cast(size_t)this.size){
        result = cast(size_t)(this.size - this.point);
        this.readEOF = true;
      }else{
        result = size;
      }
      memcpy(buffer, &this.data[cast(size_t)this.point], size);
      this.point += result;
      return(result);
    }
    unittest{
      string src = "Nageyari Software";
      ubyte[32] dest;
      std.c.stdio.printf("Unittest String_IO.read\n");
      auto io = new String_IO(cast(ubyte[])src.dup, accessible.READ);
      io.readExact(dest.ptr, 8);
      assert(dest[0..8] == "Nageyari");
      assert(io.read(dest) == src.length - 8);
      assert(dest[0..src.length-8] == " Software");
    }
    /**
     * バッファにデータを書き込む<br>
     * 残り領域以上に書き込もうとした場合，残り領域全てを書き込む
     *
     * Params:
     *  buffer = データが書き込まれる領域の先頭ポインタ
     *  size = 書き込むデータの大きさ
     * 
     * Returns:
     *  書き込めたバイト数
     * 
     * Throws:
     *  書き込みを許可していない場合，WriteException
    */
    override size_t writeBlock(const void* buffer, size_t size){
      size_t result;
      assertWriteable();
      if(size + this.point > this.size){
        //throw(new WriteException("Buffer overflow"));
        result = cast(size_t)(this.size - this.point);
      }else{
        result = size;
      }
      memcpy(&this.data[cast(size_t)this.point], buffer, result);
      this.point += result;
      return(result);
    }
    unittest{
      string src = "Nageyari Software";
      ubyte[32] dest;
      std.c.stdio.printf("Unittest String_IO.write\n");
      auto io = new String_IO(dest, accessible.WRITE);
      io.writeExact(src.ptr, 8);
      assert(dest[0..8] == "Nageyari");
      assert(io.write(cast(ubyte[])src[8..src.length]) == src.length - 8);
      assert(dest[0..src.length] == "Nageyari Software");
    }
    /**
     * バッファをシークする<br>
     * 範囲外の位置を選択した場合，範囲の端に移動する
     *
     * Params:
     *  offset = 変更した後の位置，規準位置からの相対位置で指定する
     *  whence = シークの規準位置
     * 
     * Returns:
     *  変更された現在位置
     * 
     * Throws:
     *  定義されていない規準位置を指定した場合，SeekException
    */
    override ulong seek(long offset, SeekPos whence){
      long base;
      assertSeekable();
      switch(whence){
      case SeekPos.Set:
        base = 0;
        break;
      case SeekPos.Current:
        base = this.point;
        break;
      case SeekPos.End:
        base = this.data.length;
        break;
      default:
        throw(new SeekException("Unknown seekpos"));
      }
      if(base + offset < 0){
        this.point = 0;
      }else if(base + offset > this.data.length){
        this.point = this.data.length;
      }else{
        this.point = base + offset;
      }
      this.readEOF = false;
      return(this.point);
    }
    unittest{
      std.c.stdio.printf("Unittest String_IO.seek\n");
      string src = "Nageyari Software";
      auto io = new String_IO(cast(ubyte[])src, accessible.READ);
      io.seek(3, SeekPos.Set);
      assert(io.position == 3);
      io.seek(-2, SeekPos.Current);
      assert(io.position == 1);
      io.seek(-9, SeekPos.End);
      assert(io.position == 8);
    }
    /* 解放 */
    ~this(){
    }
  }
}
